import uuid
from pathlib import Path

import radiens.grpc_radiens.common_pb2 as common_pb2
from radiens.api.api_utils.util import (path_to_source_sink_transform,
                                        source_sink_transform_to_path,
                                        to_file_ext, to_radiens_file_type)


def make_protocol_from_pb2(pb_protocol: common_pb2.Protocol):
    protocol = ProtocolAPI(pb_protocol.id)
    for pb_node in pb_protocol.transformNodes:
        protocol.add_node(make_node_from_protobuf(pb_node))
    for pb_edge in pb_protocol.transformEdges:
        protocol.add_edge(make_edge_from_protobuf(pb_edge))
    return protocol


def make_node_from_protobuf(pb_node: common_pb2.TransformNode):
    if pb_node.invalid is True:
        raise ValueError(
            'node: {} -- INVALID::{}'.format(pb_node.id, pb_node.errorMessage))
    node = TransformNode(node_id=pb_node.id)
    node._invalid = pb_node.invalid
    node._err_msg = pb_node.errorMessage
    node._pos_x = pb_node.position.x
    node._pos_y = pb_node.position.y
    if pb_node.type in [common_pb2.LOWPASS_FILTER]:
        node.lowpass(pb_node.highLowPassParams.frequency)
    elif pb_node.type in [common_pb2.HIGHPASS_FILTER]:
        node.highpass(pb_node.highLowPassParams.frequency)
    elif pb_node.type in [common_pb2.BANDPASS_FILTER]:
        node.bandpass(pb_node.bandParams.lowFrequency,
                      pb_node.bandParams.highFrequency)
    elif pb_node.type in [common_pb2.BANDSTOP_FILTER]:
        node.bandstop(pb_node.bandParams.lowFrequency,
                      pb_node.bandParams.highFrequency)
    elif pb_node.type in [common_pb2.NOTCH_FILTER]:
        node.notch(pb_node.notchParams.notchFrequency,
                   pb_node.notchParams.notchBandwidth)
    elif pb_node.type in [common_pb2.PAIRED_REF]:
        node.paired_ref(pb_node.pairedRefParams.refNtvChanIdx,
                        pb_node.pairedRefParams.targetNtvChanIdx)
    elif pb_node.type in [common_pb2.VIRTUAL_REF]:
        node.virtual_ref(pb_node.virtualRefParams.refNtvChanIdx)
    elif pb_node.type in [common_pb2.CAR_REF]:
        node.car()
    elif pb_node.type in [common_pb2.SLICE_TIME]:
        node.slice_time(pb_node.sliceTimeParams.timeStart,
                        pb_node.sliceTimeParams.timeEnd)
    elif pb_node.type in [common_pb2.SLICE_CHANNELS]:
        node.slice_channels(pb_node.sliceChannelsParams.sysChanIdxs)
    elif pb_node.type in [common_pb2.DOWNSAMPLE]:
        node.time_decimate(pb_node.downsampleParams.sampleFactor)
    elif pb_node.type in [common_pb2.SINK]:
        node.datasource_sink(Path(pb_node.datasourceSinkParams.path, pb_node.datasourceSinkParams.dsrcName +
                                  '.'+to_file_ext(pb_node.datasourceSinkParams.fileType)))
    elif pb_node.type in [common_pb2.SOURCE]:
        node.datasource_sink(Path(pb_node.datasourceSinkParams.path, pb_node.datasourceSinkParams.dsrcName +
                                  '.'+to_file_ext(pb_node.datasourceSinkParams.fileType)))
    else:
        raise ValueError('invalid transform node type {}'.format(pb_node.type))
    return node


def make_edge_from_protobuf(pb_edge: common_pb2.TransformEdge):
    return TransformEdge(edge_id=pb_edge.id, src_node_id=pb_edge.source, targ_node_id=pb_edge.target)


class TransformNode:
    def __init__(self, node_id: str):
        self._node_id = node_id
        self._node_type = None
        self._pos_x = 0
        self._pos_y = 0
        self._invalid = False
        self._err_msg = ''
        self._params = {}

    @ property
    def id(self):
        return self._node_id

    @ property
    def node_type(self):
        return self._node_type

    @ property
    def invalid(self):
        return self._invalid

    @ property
    def err_msg(self):
        return self._err_msg

    @ property
    def params(self):
        return self._params

    def lowpass(self, freq: float):
        self._node_type = common_pb2.LOWPASS_FILTER
        self._params['highLowPass'] = {'frequency': freq}
        return self

    def highpass(self, freq: float):
        self._node_type = common_pb2.HIGHPASS_FILTER
        self._params['highLowPass'] = {'frequency': freq}
        return self

    def bandpass(self, low_freq: float, high_freq: float):
        self._node_type = common_pb2.BANDPASS_FILTER
        self._params['band'] = {
            'lowFrequency': low_freq, 'highFrequency': high_freq}
        return self

    def bandstop(self, low_freq: float, high_freq: float):
        self._node_type = common_pb2.BANDSTOP_FILTER
        self._params['band'] = {
            'lowFrequency': low_freq, 'highFrequency': high_freq}
        return self

    def notch(self, freq: float, bandwidth: float):
        self._node_type = common_pb2.NOTCH_FILTER
        self._params['notch'] = {
            'notchFrequency': freq, 'notchBandwidth': bandwidth}
        return self

    def paired_ref(self, ref_ntv_chan_idx: int, target_ntv_chan_idx: int):
        self._node_type = common_pb2.PAIRED_REF
        self._params['pairedRef'] = {
            'refNtvChanIdx': ref_ntv_chan_idx, 'targetNtvChanIdx': target_ntv_chan_idx}
        return self

    def virtual_ref(self, ref_ntv_chan_idx: int):
        self._node_type = common_pb2.VIRTUAL_REF
        self._params['virtualRef'] = {'refNtvChanIdx': ref_ntv_chan_idx}
        return self

    def car(self):
        self._node_type = common_pb2.CAR_REF
        self._params['car'] = {}
        return self

    def slice_time(self, time_start: float, time_end: float):
        self._node_type = common_pb2.SLICE_TIME
        self._params['sliceTime'] = {
            'timeStart': time_start, 'timeEnd': time_end}
        return self

    def slice_channels(self, sys_chan_idxs: list):
        self._node_type = common_pb2.SLICE_CHANNELS
        self._params['sliceChannels'] = {'sysChanIdxs': sys_chan_idxs}
        return self

    def time_decimate(self, sample_factor: int):
        self._node_type = common_pb2.DOWNSAMPLE
        self._params['downsample'] = {'sampleFactor': sample_factor}
        return self

    def datasource_sink(self, path: Path):
        self._node_type = common_pb2.SINK
        self._params['datasourceSink'] = {'dsrcName': path.expanduser().resolve().stem,
                                          'path': str(path.expanduser().resolve().parent), 'fileType': to_radiens_file_type(path)}
        return self

    def datasource_source(self, path: Path):
        self._node_type = common_pb2.SOURCE
        self._params['datasourceSink'] = {'dsrcName': path.expanduser().resolve().stem,
                                          'path': str(path.expanduser().resolve().parent), 'fileType': to_radiens_file_type(path)}
        return self

    def as_protobuf(self):
        if self._node_type in [common_pb2.LOWPASS_FILTER, common_pb2.HIGHPASS_FILTER]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            highLowPassParams=common_pb2.HighLowPassTransformParams(
                                                frequency=self._params['highLowPass']['frequency']))
        if self._node_type in [common_pb2.BANDPASS_FILTER, common_pb2.BANDSTOP_FILTER]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            bandParams=common_pb2.BandTransformParams(
                                                lowFrequency=self._params['band']['lowFrequency'], highFrequency=self._params['band']['highFrequency']))
        if self._node_type in [common_pb2.NOTCH_FILTER]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            notchParams=common_pb2.NotchTransformParams(notchFrequency=self._params['notch']['notchFrequency'], notchBandwidth=self._params['notch']['notchBandwidth']))
        if self._node_type in [common_pb2.PAIRED_REF]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            pairedRefParams=common_pb2.PairedRefTransformParams(refNtvChanIdx=self._params['pairedRef']['refNtvChanIdx'], targetNtvChanIdx=self._params['pairedRef']['targetNtvChanIdx']))
        if self._node_type in [common_pb2.VIRTUAL_REF]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            virtualRefParams=common_pb2.VirtualRefTransformParams(refNtvChanIdx=self._params['virtualRef']['refNtvChanIdx']))
        if self._node_type in [common_pb2.CAR_REF]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            carParams=common_pb2.CARTransformParams())
        if self._node_type in [common_pb2.SLICE_TIME]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            sliceTimeParams=common_pb2.SliceTimeTransformParams(timeStart=self._params['sliceTime']['timeStart'], timeEnd=self._params['sliceTime']['timeEnd']))
        if self._node_type in [common_pb2.SLICE_CHANNELS]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            sliceChannelsParams=common_pb2.SliceChannelsTransformParams(sysChanIdxs=self._params['sliceChannels']['sysChanIdxs']))
        if self._node_type in [common_pb2.DOWNSAMPLE]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            downsampleParams=common_pb2.DownsampleTransformParams(sampleFactor=self._params['downsample']['sampleFactor']))
        if self._node_type in [common_pb2.SINK, common_pb2.SOURCE]:
            return common_pb2.TransformNode(id=self._node_id, type=self._node_type, invalid=False,
                                            datasourceSinkParams=common_pb2.DataSourceSinkTransformParams(dsrcName=self._params['datasourceSink']['dsrcName'], path=self._params['datasourceSink']['path'], fileType=self._params['datasourceSink']['fileType']))
        raise ValueError(
            'invalid transform node type {}'.format(self._node_type))


class TransformEdge:
    def __init__(self, edge_id: str, src_node_id: str, targ_node_id: str):
        self._edge_id = edge_id
        self._src_node_id = src_node_id
        self._targ_node_id = targ_node_id

    @ property
    def id(self):
        return self._edge_id

    @ property
    def source_node_id(self):
        return self._src_node_id

    @ property
    def target_node_id(self):
        return self._targ_node_id

    def as_protobuf(self):
        return common_pb2.TransformEdge(id=self._edge_id, source=self.source_node_id, target=self.target_node_id)


class ProtocolAPI:
    def __init__(self, protocol_id: str):
        self._protocol_id = '{}__{}'.format(
            protocol_id, str(uuid.uuid4())[4:13])
        self._transform_nodes = {}
        self._transform_edges = {}

    @ property
    def id(self):
        return self._protocol_id

    @ property
    def nodes(self):
        return self._transform_nodes

    @ property
    def edges(self):
        return self._transform_edges

    def add_node(self, node: TransformNode):
        if node.id in self._transform_nodes.keys():
            raise ValueError('{} node id already exists'.format(node.id))
        self._transform_nodes[node.id] = node
        return self

    def add_edge(self, edge: TransformEdge):
        if edge.id in self._transform_edges.keys():
            raise ValueError('{} edge id already exists'.format(edge.id))
        self._transform_edges[edge.id] = edge
        return self

    def as_protobuf(self):
        pb_edges = []
        for _, edge in self._transform_edges.items():
            pb_edges.append(edge.as_protobuf())
        pb_nodes = []
        for _, node in self._transform_nodes.items():
            pb_nodes.append(node.as_protobuf())
        return common_pb2.Protocol(id=self._protocol_id, transformNodes=pb_nodes, transformEdges=pb_edges)



from typing import TYPE_CHECKING, Union

import pandas as pd
from radiens.api import api_allego
from radiens.grpc_radiens import common_pb2, spikesorter_pb2
from radiens.lib.channel_metadata import ChannelMetadata
from radiens.lib.dataset_metadata import DatasetMetadata
from radiens.utils.constants import NEURONS1_ADDR
from radiens.utils.enums import SignalType

# the following lines are to avoid circular imports and are only used for typing hints
# (TYPE_CHECKING always evaluates to false at runtime)
if TYPE_CHECKING:
    from radiens.allego_client import AllegoClient
    from radiens.videre_client import VidereClient


class SpikeSorterClient:
    """
    Spike sorter client object for AllegoClient, VidereClient
    """

    def __init__(self, parent_client):
        """ """
        self.__parent: Union[AllegoClient, VidereClient] = parent_client

    def _is_allego(self) -> bool:
        return self.__parent.type.is_allego()

    def _is_curate(self) -> bool:
        return self.__parent.type.is_curate()

    def _is_videre(self) -> bool:
        return self.__parent.type.is_videre()

    def initialize(self):
        """
        Stops, clears, and initializes spike sorter to its default parameters.

        Returns:
            None

        Example:
            >>> client.spike_sorter().initialize()
            None

        """
        api_allego.sorter_cmd(NEURONS1_ADDR, spikesorter_pb2.SORTER_CMD_INIT)

    def rebase(self):
        """
        Rebases the spike sorter, which clears all spike data and spike templates.

        Returns:
            None

        Example:
            >>> client.spike_sorter().rebase()
            None

        """
        api_allego.sorter_cmd(NEURONS1_ADDR, spikesorter_pb2.SORTER_CMD_REBASE)

    def clear_spikes(self):
        """
        Clears all spike data from the spike sorter, but does not clear it's sorting settings or spike templates.

        Returns:
            None

        Example:
            >>> client.spike_sorter().clear_spikes()
            None

        """
        api_allego.sorter_cmd(
            NEURONS1_ADDR, spikesorter_pb2.SORTER_CMD_CLEAR_SORT)

    def set_sorting(self, mode: str):
        """
        Sets spike sorting on or off

        Parameters:
            mode (str): mode [``on``, ``off``]

        Returns:
            None

        Example:
            >>> client.spike_sorter().set_sorting('on')
            None

        """
        if mode in ["on"]:
            api_allego.sorter_cmd(NEURONS1_ADDR, spikesorter_pb2.SORTER_CMD_ON)
        elif mode in ["off"]:
            api_allego.sorter_cmd(
                NEURONS1_ADDR, spikesorter_pb2.SORTER_CMD_OFF)
        else:
            raise ValueError("mode must be in ['on', 'off']")

    def launch_offline(self, dsource: DatasetMetadata):
        pass

    def set_threshold_level(
        self,
        neg_thr=None,
        pos_thr=None,
        scale="uV",
        ntv_idxs=None,
        weak_thr=False,
        channel_meta=None,
    ):
        """
        Sets spike detection threshold levels.  This function does not effect the threshold state.

        Parameters:
            neg_thr (str): ``on`` | ``off``
            pos_thr (str): ``on`` | ``off``
            scale (str):  ``uV`` | ``sd``
            ntv_idxs (list): channel native indices (default None=all channels)
            weak_thr (bool): use True to set the weak threshold (default=False)
            channel_meta (ChannelMetadata): Allego connected channels

        Returns:
            msg (str): summary description of threshold levels

        Example:
            >>> client.spike_sorter().set_threshold_level(neg_thr=85)
            '[32 channels] detect: neg thr=85.00 uV, pos thr=n/a uV'

        See Also:
             :py:meth:`set_threshold`
             :py:meth:`get_channel_metadata`

        """
        if neg_thr in [None] and pos_thr in [None]:
            return "no new threshold levels were requested so none were changed"
        if channel_meta in [None]:
            channel_meta = self.__parent.get_channel_metadata()
        if not isinstance(channel_meta, ChannelMetadata):
            raise ValueError(
                "channel metadata must be of type ChannelMetadata")
        ntv_chan_idxs = (
            channel_meta.index(SignalType.AMP).ntv if ntv_idxs in [
                None] else ntv_idxs
        )
        reg_float = [0, 0]
        reg_bool = [False, False]
        if neg_thr not in [None]:
            reg_float[0] = neg_thr
            reg_bool[0] = True
        if pos_thr not in [None]:
            reg_float[1] = pos_thr
            reg_bool[1] = True
        if scale in ["sd"] and weak_thr in [False]:
            arg = common_pb2.SORTER_THR_LEVEL_SD
        elif scale in ["uV"] and weak_thr in [False]:
            arg = common_pb2.SORTER_THR_LEVEL
        elif scale in ["sd"] and weak_thr in [True]:
            arg = common_pb2.SORTER_WEAK_THR_LEVEL_SD
        elif scale in ["uV"] and weak_thr in [True]:
            arg = common_pb2.SORTER_WEAK_THR_LEVEL
        else:
            raise ValueError(
                'scale must be `uV` | `sd` and/or weak_thr must be True | False')
        req = [common_pb2.SpikeSorterSetParamRequest(spikeSorterID='not_used',
                                                     cmd=arg,
                                                     regFloat64=reg_float,
                                                     regBool=reg_bool,
                                                     ntvChanIdx=ntv_chan_idxs)]
        api_allego.sorter_set_params(
            NEURONS1_ADDR, common_pb2.SpikeSorterSetParamsRequest(params=req))
        return '[{} channels] {}: neg thr={} {}, pos_thr={} {}' .format(len(ntv_chan_idxs),
                                                                        'detect' if arg in [common_pb2.SORTER_THR_LEVEL,
                                                                                            common_pb2.SORTER_THR_LEVEL_SD] else 'weak',
                                                                        '{:.2f}'.format(reg_float[0]) if reg_bool[0] in [
            True] else 'n/a',
            scale,
            '{:.2f}'.format(reg_float[1]) if reg_bool[1] in [
            True] else 'n/a',
            scale)

    def set_threshold(self, neg_thr=None, pos_thr=None, ntv_idxs=None, weak_thr=False, channel_meta=None):
        """
        Set threshold 'on' or 'off'

        Parameters:
            neg_thr (str): ``on`` | ``off``
            pos_thr (str): ``on`` | ``off``
            ntv_idxs (list): channel native indices (default None=all channels)
            weak_thr (bool): use True to set the weak threshold (default=False)
            channel_meta (ChannelMetadata): Allego connected channels

        Returns:
            msg (str): summary description of threshold levels

        Example:
            >>> client.spike_sorter().set_threshold(neg_thr='on')
            '[32 channels] detect: neg thr='on', pos thr=n/a'

        See Also:
             :py:meth:`set_threshold_level`
             :py:meth:`get_channel_metadata`

        """
        if neg_thr in [None] and pos_thr in [None]:
            return "no new threshold states were requested so none were changed"
        if channel_meta in [None]:
            channel_meta = self.__parent.get_channel_metadata()
        if not isinstance(channel_meta, ChannelMetadata):
            raise ValueError(
                'channel metadata must be of type ChannelMetadata')
        ntv_idxs = channel_meta.index(SignalType.AMP).ntv
        reg_float = [0, 0]
        reg_bool = [False, False]
        if neg_thr not in [None]:
            if neg_thr not in ["on", "off"]:
                raise ValueError("neg_thr must be ``on`` or ``off``")
            reg_float[0] = 1.0 if neg_thr in ["on"] else 0.0
            reg_bool[0] = True
        if pos_thr not in [None]:
            if pos_thr not in ["on", "off"]:
                raise ValueError("pos_thr must be ``on`` or ``off``")
            reg_float[1] = 1.0 if pos_thr in ["on"] else 0.0
            reg_bool[1] = True
        if weak_thr in [False]:
            arg = common_pb2.SORTER_THR_ACTIVATE
        elif weak_thr in [True]:
            arg = common_pb2.SORTER_WEAK_THR_ACTIVATE
        else:
            raise ValueError('weak_thr must be True | False')
        req = [common_pb2.SpikeSorterSetParamRequest(spikeSorterID='not_used',
                                                     cmd=common_pb2.arg,
                                                     regFloat64=reg_float,
                                                     regBool=reg_bool,
                                                     ntvChanIdx=ntv_idxs)]
        api_allego.sorter_set_params(
            NEURONS1_ADDR, common_pb2.SpikeSorterSetParamsRequest(params=req))
        return '[{} channels] {}: neg thr={}, pos_thr={}' .format(len(ntv_idxs),
                                                                  'detect' if arg in [common_pb2.SORTER_THR_LEVEL,
                                                                                      common_pb2.SORTER_THR_LEVEL_SD] else 'weak',
                                                                  '{}'.format(neg_thr) if reg_bool[0] in [
            True] else 'n/a',
            '{}'.format(pos_thr) if reg_bool[1] in [True] else 'n/a')

    def set_spike_window(self, pre_thr_ms=None, post_thr_ms=None):
        """
        Set spike window.

        Parameters:
            pre_thr_ms (float): pre-threshold window duration in milliseconds (default None=not changed)
            post_thr_ms (float): post-threshold window duration in milliseconds (default None=not changed)

        Returns:
            msg (str): summary description of spike window

        Example:
            >>> client.spike_sorter().set_threshold(neg_thr='on')
            '[32 channels] detect: neg thr='on', pos thr=n/a'

        See Also:
             :py:meth:`set_threshold_level`
             :py:meth:`get_channel_metadata`

        """
        ntv_chan_idxs = self.__parent.get_channel_metadata().index(SignalType.AMP).ntv
        params = self.get_params()
        if len(params) == 0:
            return "no connected channels, spike window was not changed"
        cur_wdw = params.loc[params["ntv_idx"] == 0]["thr_wdw"][0]
        reg_float = [pre_thr_ms, post_thr_ms]
        reg_bool = [True, True]
        if pre_thr_ms in [None]:
            reg_float[0] = cur_wdw[0] * 0.001  # converts to sec
        if post_thr_ms in [None]:
            reg_float[1] = cur_wdw[1] * 0.001  # converts to sec
        req = [common_pb2.SpikeSorterSetParamRequest(spikeSorterID='not_used',
                                                     cmd=common_pb2.SORTER_THR_WDW,
                                                     regFloat64=reg_float,
                                                     regBool=reg_bool,
                                                     ntvChanIdx=ntv_chan_idxs)]
        api_allego.sorter_set_params(
            NEURONS1_ADDR, common_pb2.SpikeSorterSetParamsRequest(params=req))
        params = self.get_params()
        cur_wdw = params.loc[params['ntv_idx'] == 0]['thr_wdw'][0]
        cur_pts = params.loc[params['ntv_idx'] == 0]['thr_wdw_pts'][0]
        return '[{} channels] window: pre-thr={:.2f}, post-thr={:.2f}, total={:.2f} ms [{} points]' .format(len(ntv_chan_idxs),
                                                                                                            cur_wdw[0]*1000.0, cur_wdw[1] *
                                                                                                            1000.0, cur_wdw[2] *
                                                                                                            1000.0,
                                                                                                            cur_pts[2])

    def set_spike_shadow(self, shadow_ms=None):
        """
        Sets spike shadow window.  The shadow window is the time skipped after each threshold crossing.

        Parameters:
            shadow_ms (float): shadow window duration in ms (default None=not changed)

        Returns:
            msg (str): summary description of spike shadow window

        Example:
            >>> client.spike_sorter().set_spike_shadow(1.5)
            '[32 channels] shadow window: 1.5 ms [40 points]'

        See Also:
             :py:meth:`set_threshold_level`
             :py:meth:`get_channel_metadata`

        """
        ntv_chan_idxs = self.__parent.get_channel_metadata().index(SignalType.AMP).ntv
        params = self.get_params()
        if len(params) == 0:
            return "no connected channels, spike shadow window was not changed"
        cur_wdw = params.loc[params["ntv_idx"] == 0]["shadow"][0]
        reg_float = [shadow_ms, 0.0]
        reg_bool = [True, True]
        if shadow_ms in [None]:
            reg_float[0] = cur_wdw[0] * 0.001  # converts to sec
        req = [common_pb2.SpikeSorterSetParamRequest(spikeSorterID='not_used',
                                                     cmd=common_pb2.SORTER_SHADOW,
                                                     regFloat64=reg_float,
                                                     regBool=reg_bool,
                                                     ntvChanIdx=ntv_chan_idxs)]
        api_allego.sorter_set_params(
            NEURONS1_ADDR, common_pb2.SpikeSorterSetParamsRequest(params=req))
        params = self.get_params()
        cur_wdw = params.loc[params["ntv_idx"] == 0]["shadow"][0]
        cur_pts = params.loc[params["ntv_idx"] == 0]["shadow_pts"][0]
        return "[{} channels] shadow window: {:.2f}  ms [{} points]".format(
            len(ntv_chan_idxs), cur_wdw * 1000.0, cur_pts
        )

    def get_params(self) -> pd.DataFrame:
        """
        Returns the spike sorter parameters as a table.

        Returns:
            pandas.DataFrame

        Example:
            >>> client.sorter_get_params()
            None

        """
        return api_allego.sorter_get_params(NEURONS1_ADDR)

    def get_state(self) -> pd.DataFrame:
        """
        Returns the spike sorter state as a table.

        Returns:
            pandas.DataFrame

        Example:
            >>> client.sorter_get_state()
            None

        """
        return api_allego.sorter_get_state(NEURONS1_ADDR)

    def get_dashboard(self) -> pd.DataFrame:
        """
        Returns the spike sorter dashboard as a table.

        Returns:
            pandas.DataFrame

        Example:
            >>> client.sorter_get_dashboard()
            None

        """
        return api_allego.sorter_get_dashboard(NEURONS1_ADDR)



from __future__ import annotations

import math
import warnings
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import radiens.api.api_allego as api_allego
import radiens.api.api_videre as api_videre
from radiens.grpc_radiens import common_pb2
from radiens.lib.dataset_metadata import DatasetMetadata
from radiens.lib.signals_snapshot import PSD, Signals
from radiens.utils.constants import DEFAULT_HUB_ID, SigSelect, TimeRange
from radiens.utils.enums import (FftWindow, KeyIndex, PsdScaling,
                                 RadiensService, SignalType, TrsMode)
from radiens.utils.util import (make_time_range, sig_sel_to_protobuf,
                                time_range_to_protobuf)

# the following lines are to avoid circular imports and are only used for typing hints
# (TYPE_CHECKING always evaluates to false at runtime)
if TYPE_CHECKING:
    from radiens.videre_client import VidereClient


class SignalsClient:
    """
    Signals client object for Allego, Curate, and Videre
    """

    def __init__(self, parent_client: VidereClient):
        """
        """
        if not parent_client.type.is_videre:
            raise ValueError(
                "signals_client: parent_client must be a VidereClient")
        self.__parent: VidereClient = parent_client

    def get_signals(self,
                    time_range: TimeRange | list[float] | np.ndarray = None,
                    sel_mode: TrsMode = None,
                    sig_sel: SigSelect = None,
                    dataset_metadata: DatasetMetadata = None,
                    hub_name=DEFAULT_HUB_ID) -> Signals:
        """
        Gets signals over the specified time range from a linked dataset.

        Parameters:
            time_range (TimeRange): see :py:meth:`~radiens.utils.util.make_time_range`
            dataset_metadata: see :py:meth:`link_data_file` and :py:meth:`get_data_file_metadata`
            sel_mode: optional (default :py:attr:`TrsMode.SUBSET`)
            sig_sel: optional
            hub_name: optional


        Returns:
            signals (Signals)
        """
        if dataset_metadata is None:
            raise ValueError('videre: dataset_metadata must be provided')
        if time_range is None:
            warnings.warn(
                f'no requested time range, using dataset time range {dataset_metadata.TR.sec} (s) ')
            time_range = dataset_metadata.TR
        if not isinstance(time_range, TimeRange):
            if not isinstance(time_range, (list, np.ndarray)):
                raise ValueError(
                    'time_range must be a TimeRange, list, or numpy array')
            time_range = make_time_range(
                time_range=time_range, fs=dataset_metadata.TR.fs)
        if sel_mode is None:
            sel_mode = TrsMode.SUBSET
        if sig_sel is None:
            sig_sel = SigSelect(key_idx=KeyIndex.NTV,
                                amp=np.array(dataset_metadata.channel_metadata.index(
                                    SignalType.AMP).ntv, dtype=np.int64),
                                gpio_ain=np.array(dataset_metadata.channel_metadata.index(
                                    SignalType.AIN).ntv, dtype=np.int64),
                                gpio_din=np.array(dataset_metadata.channel_metadata.index(
                                    SignalType.DIN).ntv, dtype=np.int64),
                                gpio_dout=np.array(dataset_metadata.channel_metadata.index(
                                    SignalType.DOUT).ntv, dtype=np.int64),
                                )
        elif sig_sel.key_idx != KeyIndex.NTV:
            raise ValueError("sig_sel must use KeyIndex.NTV as the key index")

        sig_sel_proto = sig_sel_to_protobuf(sig_sel)
        req = common_pb2.HDSnapshotRequest2(
            dsourceID=dataset_metadata.id,
            selMode=sel_mode.value,
            sigSel=sig_sel_proto,
            tR=time_range_to_protobuf(time_range)
        )

        return api_videre.get_signals(
            self.__parent._server_address(hub_name, RadiensService.CORE), req)

    def get_psd(
        self,
        time_range: TimeRange | list[float] | np.ndarray = None,
        sel_mode: TrsMode = None,
        sig_sel: SigSelect = None,
        samp_freq: float = 2000.0,
        freq_range: list | np.ndarray = [1, 300],
        collapse_freq: bool = False,
        scaling: PsdScaling = PsdScaling.ABSOLUTE,
        window: FftWindow = FftWindow.HAMMING_p01,
        freq_resolution: float = None,
        file: str = None,
        dataset_metadata: DatasetMetadata = None,
        hub_name=DEFAULT_HUB_ID,
    ) -> PSD:
        """
        Gets the power spectral density (PSD) for the specified signals over the specified time range.

        Parameters:
            time_range (TimeRange) : requested time range in seconds (required)
            sel_mode (TrsMode): time range selection mode (optional, default=TrsMode.SUBSET)
            sig_sel (SigSelect): requested AMP signals (optional, default=all AMP signals)
            samp_freq (float): PSD sample frequency in Hz. (optional, default=dataset_metadata.TR.fs)
            freq_range (list, np.ndarray): requested frequency range.  (optional, default=[1, 300])
            collapse_freq (bool): True collapses `freq_range` into one frequency bin (optional, default=False)
            scaling (PsdScaling): sets the PSD scale (optional, default=PsdScaling.ABSOLUTE)
            window (FftWindow): sets the FFT window(optional, default=FftWindow.HAMMING_p01)
            freq_resolution (float): requested frequency resolution in Hz (optional, default=None)
            file (string): save psd data to file (optional, default=None)
            dataset_metadata (DatasetMetadata): dataset metadata (required for Videre, optional for Allego, default=None)
            hub_name (str): Radiens hub name (optional, default=DEFAULT_HUB)


        Returns:
            psd (PSD): container object for PSD data.
        """
        if dataset_metadata is None:
            raise ValueError('videre: dataset_metadata must be provided')
        if time_range is None:
            warnings.warn(
                f'no requested time range, using dataset time range {dataset_metadata.TR.sec} (s) ')
            time_range = dataset_metadata.TR
        if not isinstance(time_range, TimeRange):
            time_range = make_time_range(
                time_range=time_range, fs=dataset_metadata.TR.fs)
        if sel_mode is None:
            sel_mode = TrsMode.SUBSET
        if sig_sel is None:
            sig_sel = SigSelect(key_idx=KeyIndex.NTV,
                                amp=np.array(dataset_metadata.channel_metadata.index(
                                    SignalType.AMP).ntv, dtype=np.int64),
                                gpio_ain=np.array(dataset_metadata.channel_metadata.index(
                                    SignalType.AIN).ntv, dtype=np.int64),
                                gpio_din=np.array(dataset_metadata.channel_metadata.index(
                                    SignalType.DIN).ntv, dtype=np.int64),
                                gpio_dout=np.array(dataset_metadata.channel_metadata.index(
                                    SignalType.DOUT).ntv, dtype=np.int64),
                                )
        elif sig_sel.key_idx != KeyIndex.NTV:
            raise ValueError("sig_sel must use KeyIndex.NTV as the key index")

        samp_freq = time_range.fs if samp_freq is None else samp_freq

        req = common_pb2.PSDRequest(
            tR=time_range_to_protobuf(time_range),
            selMode=int(sel_mode.value),
            ntvIdxs=sig_sel.amp,
            stype=SignalType.AMP.value,
            resampleFs=samp_freq,
            wdwType=window.value,
            scaling=scaling.value,
            freqRange=freq_range,
            deltaFreq=freq_resolution,
            collapseFreq=collapse_freq,
            path=file,
            isReturnPSD=True,
            dsourceID=dataset_metadata.id,
        )
        out = api_videre.get_psd(
            self.__parent._server_address(
                hub_name, RadiensService.CORE), req
        )
        if file is not None:
            out.save(file)

        return out



from collections.abc import Iterable
from pathlib import Path

import numpy as np
import pandas as pd
from radiens.api import api_curate, api_videre
from radiens.api.api_utils.util import (BaseClient, get_dataset_id,
                                        to_radiens_file_type)
from radiens.grpc_radiens import common_pb2, datasource_pb2, radiensserver_pb2
from radiens.lib.dataset_metadata import DatasetMetadata
from radiens.lib.summa import SummaAnalysis, SummaStatus
from radiens.metrics_client import MetricsClient
from radiens.signals_client import SignalsClient
from radiens.spike_sorter_client import SpikeSorterClient
from radiens.spikes_client import SpikesClient
from radiens.utils.constants import DEFAULT_HUB_ID, TimeRange
from radiens.utils.enums import ClientType, RadiensService
from radiens.utils.util import make_time_range


class VidereClient(BaseClient):
    """
    VidereClient implements the radiens API for offline analysis and visualization.
    It matches and extends the functionality of the Radiens Videre UI app.
    """

    def __init__(self, hub_name=DEFAULT_HUB_ID):
        """
        """
        super().__init__()
        self._default_hub = hub_name
        self._sorter = SpikeSorterClient(self)
        self._spikes = SpikesClient(self)
        self._metrics = MetricsClient(self)
        self._signals = SignalsClient(self)
        self._meta_cache = {}

    @property
    def hubs(self) -> dict:
        """
        dict of active radiens hubs, with hub ID as key.
        """
        return self._hubs()

    @property
    def hub(self) -> str:
        """
        Radiens hub for this instance of VidereClient.
        """
        return self._default_hub

    @property
    def id(self) -> str:
        """
        UID of this client session.
        """
        return self._id()

    @property
    def type(self) -> ClientType:
        """
        Returns :py:attr:`~radiens.utils.enums.ClientType.VIDERE`
        """
        return ClientType.VIDERE

    def signal_metrics(self) -> MetricsClient:
        """
        Videre signal metrics API
        """
        return self._metrics

    def spike_sorter(self) -> SpikeSorterClient:
        """
        Videre spike sorter API
        """
        return self._sorter

    def spikes(self) -> SpikesClient:
        """
        Videre spikes API
        """
        return self._spikes

    def signals(self) -> SignalsClient:
        """
        Signals API
        """
        return self._signals

    def _cache_file_meta(self, meta: DatasetMetadata):
        self._meta_cache[meta.id] = meta

    def _uncache_file_meta(self, id: str | DatasetMetadata):
        if isinstance(id, DatasetMetadata):
            id = id.id
        self._meta_cache.pop(id, None)

    def link_data_file(self, path:  str | Path, calc_metrics: bool = True, force: bool = False, hub_name=DEFAULT_HUB_ID, link_spikes: bool = True) -> DatasetMetadata:
        """
            Links a Radiens data file to the Radiens hub and returns the data set meta data.
            If force=False (default) and the 'path' is already linked then nothing is done and existing data set meta data is returned.
            Use force=True to always link the source as a new dataset. All instances of a linked data set are backed by the same data files.

            Parameters:
                path (str, pathlib.Path): path to source file
                calc_metrics (bool) optional: use True to calculate signal metrics (default=True)
                force (bool) optional: use True to force the source to be linked to the hub (default=False)
                hub_name (str) optional: radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

            Returns:
                data_source (DatasetMetadata): data file metadata

            See Also:
                :py:meth:`rename_file`
                :py:meth:`clear_dataset()`
                :py:meth:`get_dataset_ids()`
        """
        _ftype = to_radiens_file_type(path)  # defaults to xdat
        path = Path(path).expanduser().resolve()
        _ext = path.suffix.lower()
        if _ext == '':
            data_sets = api_curate.dsrc_list_dir(self._server_address(
                hub_name, RadiensService.CORE), [str(path)], sort_by="type")
            if len(data_sets.datasources) == 1:
                # 'x' is a dummy stem so to_radiens_file_type will return the correct file type
                _ext = 'x.'+data_sets.datasources[0].descriptor.file_type
                _ftype = to_radiens_file_type(Path(_ext))
        req = datasource_pb2.DataSourceSetSaveRequest(
            path=str(path.parent),
            baseName=path.stem,
            fileType=_ftype.value,
            dsourceID=path.stem + self.id,
            isBackgroundKPI=calc_metrics,
            isForce=force,
            forceNoSorter=True,
        )
        dataset_meta, associated_spikes = api_videre.set_datasource(
            self._server_address(hub_name, RadiensService.CORE), req)
        self._cache_file_meta(dataset_meta)

        if link_spikes:
            for spike_path in associated_spikes:
                req = datasource_pb2.DataSourceSetSaveRequest(
                    path=str(spike_path.parent),
                    baseName=spike_path.stem,
                    fileType=to_radiens_file_type(spike_path).value,
                    dsourceID=spike_path.stem,
                    isBackgroundKPI=False,
                    isForce=force,
                )
                spike_meta, _ = api_videre.set_datasource(
                    self._server_address(hub_name, RadiensService.CORE), req)
                self._cache_file_meta(spike_meta)
        return dataset_meta

    def _parse_time_range(self, dataset_id: str, time_range: TimeRange | list = None) -> TimeRange:
        if isinstance(time_range, TimeRange):
            return time_range
        else:
            meta = self._get_file_meta_from_id(dataset_id)
            if not isinstance(time_range, list):
                return meta.time_range
            else:
                return make_time_range(time_range, fs=meta.time_range.fs)

    def get_data_file_metadata(self, dataset_idx: int = None, dataset_id: str = None, path: str | Path = None, hub_name=DEFAULT_HUB_ID, fail_hard=False) -> DatasetMetadata:
        """
            Returns meta data for one Radiens dataset or file.  The dataset can be requested by index, ID or full path, in that order of preference.

            Parameters:
                dataset_idx (int): index of dataset in table returned by :py:meth:`get_dataset_ids`
                dataset_id (str): dataset ID as listed in table returned by :py:meth:`get_dataset_ids` 
                path (str, pathlib.Path): full path to data file set in form of `path/base_name.ext`.
                hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)
                fail_hard (bool): if True, then an exception is raised if the requested dataset is not available.

            Returns:
                dataset_meta (DatasetMetadata): data set metadata

            Side Effects:
                If the requested dataset is specified by `path`, then it is linked to the hub and remains linked.

            See Also:
                :py:meth:`rename_file`
                :py:meth:`_clear_dataset`
                :py:meth:`get_dataset_ids`
        """
        dataset_id, path = get_dataset_id(self.get_dataset_ids(), self.id, hub_name, dataset_idx=dataset_idx,
                                          dataset_id=dataset_id, path=path, fail_hard=fail_hard)
        if dataset_id is not None:
            if path is None:
                return self._get_file_meta_from_id(dataset_id, force_refresh=True, hub_name=hub_name)
            else:
                req = datasource_pb2.DataSourceSetSaveRequest(path=str(path.expanduser().resolve().parent), baseName=path.stem, fileType=to_radiens_file_type(
                    path).value, isBackgroundKPI=False, isForce=False)
                return api_videre.set_datasource(self._server_address(hub_name, RadiensService.CORE), req)
        raise ValueError('No dataset found')

    def _get_file_meta_from_id(self, dataset_id: str, force_refresh=False, hub_name=DEFAULT_HUB_ID) -> DatasetMetadata:
        if not force_refresh and dataset_id in self._meta_cache:
            return self._meta_cache[dataset_id]
        else:
            req = datasource_pb2.DataSourceSetSaveRequest(
                dsourceID=dataset_id,
                isBackgroundKPI=False,
                isForce=False,
            )
            meta, _ = api_videre.set_datasource(
                self._server_address(hub_name, RadiensService.CORE), req)
            self._cache_file_meta(meta)
            return meta

    def clear_dataset(self, dataset_idx: int = None, dataset_id: str = None, path: str | Path = None, hub_name=DEFAULT_HUB_ID, fail_hard=False) -> any:
        """
            Clears one or more datasets of this Curate session from the Radiens hub.
            Use dataset_id='all' to clear all session datasets.
            This is a power user function.

            Parameters:
                source (str, list) optional: data source file or list of data source files.
                dataset_id (str, list) optional: dataset ID or list of dataset IDs (default=[], 'all' clears all datasets for the Curate session)
                hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

            Returns:
                num_unlinked (int): number of unlinked datasets
                dataset_ids (pandas.DataFrame): linked dataset IDs on the Radiens hub.

            See Also:
                :py:meth:`link_data_file()`
                :py:meth:`get_dataset_ids()`
        """
        available_dataset_ids = self.get_dataset_ids()
        if dataset_id == 'all' or dataset_idx == 'all':
            dataset_id = available_dataset_ids['dataset ID'].values
            api_videre.unlink_datasource(self._server_address(
                hub_name, RadiensService.CORE), dataset_id)
            for dset_id in dataset_id:
                self._uncache_file_meta(dset_id)
        else:
            dataset_id, _ = get_dataset_id(available_dataset_ids, self.id, hub_name, dataset_idx=dataset_idx,
                                           dataset_id=dataset_id, path=path, fail_hard=fail_hard)
            if dataset_id is not None:
                api_videre.unlink_datasource(self._server_address(
                    hub_name, RadiensService.CORE), [dataset_id])
                self._uncache_file_meta(dataset_id)
        new_avail = self.get_dataset_ids(hub_name=hub_name)
        return len(available_dataset_ids)-len(new_avail), new_avail

    def get_dataset_ids(self, hub_name=DEFAULT_HUB_ID) -> pd.DataFrame:
        """
            Returns sorted list of linked dataset IDs on the Radiens hub.

            Returns:
                dataset_ids (pandas.DataFrame): linked dataset IDs on the Radiens hub.

            See Also:
                :py:meth:`link_data_file()`
                :py:meth:`clear_dataset()`
        """
        resp = api_videre.list_datasource_ids(
            self._server_address(hub_name, RadiensService.CORE))
        df = pd.DataFrame({'dataset ID': resp, 'base name': [x[:x.find('_-_')]
                                                             for x in resp]}).sort_values(by='dataset ID', axis=0).reset_index(drop=True)
        df.index.name = 'index'
        return df

    def dashboard(self, close: bool = False, hub_name=DEFAULT_HUB_ID) -> None:
        """
            Launches a dashboard for the Radiens Hub

            Returns:
                None
        """
        args = common_pb2.DashboardCommandRequest.Args(
            cmd=common_pb2.DASH_CMD_CLOSE) if close else common_pb2.DashboardCommandRequest.Args(cmd=common_pb2.DASH_CMD_OPEN)
        req = common_pb2.DashboardCommandRequest(
            args=args, dashType=common_pb2.DASH_1)
        api_videre.command_dashboard(
            self._server_address(hub_name, RadiensService.DASH), req)

    def summa_list_sessions(self, hub_name=DEFAULT_HUB_ID) -> None:
        """
            Lists the current Summa analysis sessions for the Radiens Hub

            Returns:
                None
        """
        return api_videre.summa_list_sessions(self._server_address(hub_name, RadiensService.DASH), common_pb2.StandardRequest())

    def summa_clear_sessions(self, hub_name=DEFAULT_HUB_ID) -> None:
        """
            Lists the current Summa analysis sessions for the Radiens Hub

            Returns:
                None
        """
        return api_videre.summa_clear_sessions(self._server_address(hub_name, RadiensService.DASH), common_pb2.StandardRequest())

    def summa_launch_session_analysis(self, summa_id: str = None, dataset_ids: list = None, bin_dur_sec: float = 1.0,  hub_name=DEFAULT_HUB_ID) -> None:
        """
            Launches a new Summa analysis session.

            Returns:
                status (SummaStatus): Summa analysis status.
        """
        if not isinstance(dataset_ids, list):
            raise ValueError('dataset_ids must be a list')
        req = radiensserver_pb2.SummaAnalysisRequest(
            summaID='default' if summa_id is None else summa_id,
            dsourceIDs=dataset_ids,
            pktDurSec=bin_dur_sec,
        )
        return SummaAnalysis(api_videre.summa_launch_session_analysis(self._server_address(hub_name, RadiensService.DASH), req)).status

    def summa_session_status(self, summa_id: str = None, hub_name=DEFAULT_HUB_ID) -> None:
        """
            Launches a dashboard for the Radiens Hub

            Returns:
                None
        """
        req = radiensserver_pb2.SummaAnalysisRequest(
            summaID='default' if summa_id is None else summa_id,
            reqPart='',
        )
        resp = api_videre.summa_get_session_analysis(
            self._server_address(hub_name, RadiensService.DASH), req)
        return SummaStatus(resp.state)

    def summa_get_session_aggregate_statistics(self, summa_id: str = None, hub_name=DEFAULT_HUB_ID) -> None:
        """
            Launches a dashboard for the Radiens Hub

            Returns:
                None
        """
        req = radiensserver_pb2.SummaAnalysisRequest(
            summaID='default' if summa_id is None else summa_id,
            reqPart='agg',
        )
        return SummaAnalysis(api_videre.summa_get_session_analysis(self._server_address(hub_name, RadiensService.DASH), req))

    def summa_get_session_datasource_statistics(self, summa_id: str = None, dataset_ids: list = None, hub_name=DEFAULT_HUB_ID) -> None:
        """
            Launches a dashboard for the Radiens Hub

            Returns:
                None
        """
        dsrc_part = 'dsource:'
        if dataset_ids is None or len(dataset_ids) == 0:
            dsrc_part += 'all'
        else:
            for dset_id in dataset_ids:
                dsrc_part += '{},'.format(dset_id)
        req = radiensserver_pb2.SummaAnalysisRequest(
            summaID='default' if summa_id is None else summa_id,
            reqPart=dsrc_part,
        )
        return SummaAnalysis(api_videre.summa_get_session_analysis(self._server_address(hub_name, RadiensService.DASH), req))



from collections import OrderedDict, namedtuple

import numpy as np
import pandas as pd
from radiens.grpc_radiens import radiensserver_pb2
from radiens.lib.channel_metadata import ChannelMetadata
from radiens.lib.spikes import SpikesMetadata
from radiens.utils.enums import MetricID, MetricMode, MetricName
from radiens.utils.util import (Histogram, from_dense_matrix,
                                is_protobuf_time_range_empty, make_time_range)


class SummaStatus:
    """
    Container for Summa analysis status
    """

    def __init__(self, msg: radiensserver_pb2.SummaState):
        if not isinstance(msg, radiensserver_pb2.SummaState):
            raise ValueError("status wrong type")
        self._d = {
            "summa_id": msg.summaID,
            "is_complete": msg.isComplete,
            "frac_complete": msg.fracComplete,
            "msg": msg.msg,
            "dsource_walltime": {
                "start": msg.dsrcWalltimeStart,
                "end": msg.dsrcWalltimeEnd,
            },
            "analysis_walltime_start": msg.analysisWallTimeStart,
            "analysis_elapsed_time_sec": msg.analysisElapsedSec,
            "packet_dur_sec": msg.packetDurSec,
        }

    @property
    def id(self) -> str:
        """
        Signal metrics time range of either the backing file (if available) or the backing cache.
        """
        return self._d["summa_id"]

    @property
    def is_complete(self) -> bool:
        """
        Signal metrics time range of either the backing file (if available) or the backing cache.
        """
        return self._d["is_complete"]

    @property
    def frac_complete(self) -> float:
        """
        Signal metrics time range of either the backing file (if available) or the backing cache.
        """
        return self._d["frac_complete"]

    @property
    def packet_dur_sec(self) -> float:
        """
        Signal metrics time range of either the backing file (if available) or the backing cache.
        """
        return self._d["packet_dur_sec"]

    def string(self) -> str:
        return "SUMMA:{} - started @ {} (duration: {:.3f} sec); {} ({:.2f}/1.00), packet dur={:.3f} sec".format(
            self.id,
            self._d["analysis_walltime_start"],
            self._d["analysis_elapsed_time_sec"],
            "complete" if self.is_complete else "working",
            self.frac_complete,
            self.packet_dur_sec,
        )


class SummaAggregateStats:
    """
    Container for Summa analysis status
    """

    def __init__(self, msg: radiensserver_pb2.SummaAggregateStats):
        if not isinstance(msg, radiensserver_pb2.SummaAggregateStats):
            raise ValueError("msg wrong type")

        self._metric_ids = []
        for m_id in msg.metricID:
            self._metric_ids.append(
                MetricID(mode=MetricMode(m_id.mode),
                         name=MetricName(m_id.name))
            )
        self._hist = OrderedDict()
        for k, hist in enumerate(msg.hist):
            self._hist[self._metric_ids[k]] = Histogram(hist)

    @property
    def metric_ids(self) -> list:
        """
        Signal metrics time range of either the backing file (if available) or the backing cache.
        """
        return self._metric_ids

    @property
    def histogram(self) -> OrderedDict:
        """
        Signa metric histograms aggregated over all sites and all requested datasources.
        """
        return self._hist


class SummaDatasourceStats:
    """
    Container for Summa analysis status
    """

    def __init__(self, msg: radiensserver_pb2.SummaDatasourceAnalysisPkg):
        if not isinstance(msg, radiensserver_pb2.SummaDatasourceAnalysisPkg):
            raise ValueError("msg wrong type")
        self._ntv_idxs = np.array(msg.ntvIdxs, dtype=np.int64)
        self._hist = []
        for hist in msg.hist:
            self._hist.append(Histogram(hist))
        self._recs = []
        for rec in msg.recs:
            self._recs.append(from_dense_matrix(rec))
        self._hist_recs = []
        for hist in self._hist_recs:
            self._hist_recs.append(Histogram(hist))
        self._metric_ids = []
        for m_id in msg.metricID:
            self._metric_ids.append(
                MetricID(mode=MetricMode(m_id.mode),
                         name=MetricName(m_id.name))
            )

    @property
    def metric_ids(self) -> list:
        """
        Signal metrics time range of either the backing file (if available) or the backing cache.
        """
        return self._metric_ids

    @property
    def histogram(self) -> Histogram:
        """
        Signal metrics time range of either the backing file (if available) or the backing cache.
        """
        return self._hist


class SummaAnalysis:
    """
    Container for Summa signal analysis
    """

    def __init__(self, msg: radiensserver_pb2.SummaAnalysisReply):
        if not isinstance(msg, radiensserver_pb2.SummaAnalysisReply):
            raise ValueError("message wrong type")
        self._d = {
            "summa_id": msg.summaID,
            "all_dsource_ids": msg.allDsourceIDs,
            "req_available_dsource_ids": msg.reqAvailDsourceIDs,
            "status": SummaStatus(msg.state),
        }
        self._agg_stats = SummaAggregateStats(msg.aggStats)

        if len(self._d["all_dsource_ids"]) != len(msg.allDsourceSpec):
            raise ValueError("inconsistent number of datasource specs")
        self._dsrc_spec = {}
        for k, spec in enumerate(msg.allDsourceSpec):
            x = {"SG": ChannelMetadata(
                spec.sG), "TR": make_time_range(pbTR=spec.tR)}

            spec.niSpec.tR is None if is_protobuf_time_range_empty(
                spec.niSpec.tR
            ) else SpikesMetadata(spec.niSpec)
            self._dsrc_spec[self._d["all_dsource_ids"][k]] = x

        if len(self._d["req_available_dsource_ids"]) != len(msg.reqAvailDsourceStats):
            raise ValueError("inconsistent number of datasource stats")
        self._dsrc_stats = {}
        for k, stats in enumerate(msg.reqAvailDsourceStats):
            self._dsrc_stats[
                self._d["req_available_dsource_ids"][k]
            ] = SummaDatasourceStats(stats)

    @property
    def id(self) -> str:
        """
        Summa analysis ID
        """
        return self._d["summa_id"]

    @property
    def status(self) -> SummaStatus:
        """
        Status of the Summa analysis session.
        """
        return self._d["status"]

    @property
    def all_dsource_info(self) -> dict:
        """
        Summary information on all datasources included in this Summa analysis.
        """
        return self._dsrc_spec

    @property
    def aggregate_stats(self) -> SummaAggregateStats:
        """
        Summary statistics aggregated over the requested datasources.
        """
        return self._agg_stats

    @property
    def datasource_stats(self) -> dict:
        """
        Summary statistics on each of the requested datasources.
        """
        return self._dsrc_stats



import time
import uuid
import warnings
from collections.abc import Iterable
from pathlib import Path

import radiens.api.api_curate as api_curate
import radiens.api.api_videre as api_videre
from grpc import RpcError
from radiens.api.api_utils.protocols import (ProtocolAPI, TransformEdge,
                                             TransformNode)
from radiens.api.api_utils.util import (BaseClient, ext_to_radiens_file_type,
                                        to_file_ext, to_radiens_file_type,
                                        to_suffix)
from radiens.exceptions.grpc_error import RpcException
from radiens.file_sys_client import FileSystemClient
from radiens.grpc_radiens import datasource_pb2
from radiens.lib.dataset_metadata import DatasetMetadata
from radiens.metrics_client import MetricsClient
from radiens.spikes_client import SpikesClient
from radiens.utils.constants import (CONVERTIBLE_RADIENS_FILE_TYPES,
                                     DEFAULT_HUB_ID)
from radiens.utils.enums import ClientType, RadiensFileType, RadiensService
from radiens.utils.util import rm_xdat_file
from tqdm.autonotebook import tqdm


class CurateClient(BaseClient):
    """
    CurateClient implements the radiens API for data curation. It matches and extends the functionality of the Radiens Curate UI app.
    """

    def __init__(self):
        """ """
        super().__init__()
        self._spikes = SpikesClient(self)
        self._metrics = MetricsClient(self)
        self._fsys = FileSystemClient()
        self._dsp_stream = {}

    @property
    def hubs(self) -> dict:
        """
        dict of active radiens hubs, with hub ID as key.
        """
        return self._hubs()

    @property
    def id(self) -> str:
        """
        UID of this client session.
        """
        return self._id()

    @property
    def type(self) -> ClientType:
        """
        Returns :py:attr:`~radiens.utils.enums.ClientType.CURATE`
        """
        return ClientType.CURATE

    def signal_metrics(self) -> MetricsClient:
        """
        Signal metrics API
        """
        return self._metrics

    def spikes(self) -> SpikesClient:
        """
        Spikes API
        """
        return self._spikes

    @property
    def dsp_stream(self):
        """
        dict of DSP streams running in the background.
        """
        return self._dsp_stream

    def dsp_clean_up(self):
        """
        Cleans up after any DSP operation that has finished.

        Returns:
            None

        Examples:
            >>> stream_id = client.dsp_low_pass_filter(750, './my_source_file', './my_output_file')
            >>> progress_bar(stream_id)

        See Also:
            :py:meth:`dsp_progress`
            :py:meth:`dsp_low_pass_filter`

        """
        to_del = []
        for stream_id in self.dsp_stream:
            if self.dsp_stream[stream_id]["stream"].done():
                to_del.append(stream_id)
        for stream_id in to_del:
            del self.dsp_stream[stream_id]

    def progress_bar(self, stream_id):
        """
        Displays a progress bar showing the progress of the DSP stream running in the background.
        This function blocks until the background DSP stream is complete.
        This is an optional function that does not effect the DSP operation.

        Parameters:
            stream_id (str): stream ID of the requested DSP stream.

        Returns:
            None

        Examples:
            >>> stream_id = client.dsp_low_pass_filter(750, './my_source_file', './my_output_file')
            >>> progress_bar(stream_id)

        See Also:
            :py:meth:`dsp_progress`
            :py:meth:`dsp_low_pass_filter`

        """
        pbar = tqdm(
            total=self.dsp_stream[stream_id]["data_source"].time_range.dur_sec,
            desc=stream_id,
            smoothing=0.9,
            bar_format="{desc}: {percentage:.1f}%|{bar}| {n:.2f}/{total_fmt} [{elapsed}<{remaining}]",
        )
        for progress in self.dsp_stream[stream_id]["stream"]:
            if progress.incrementSec is None:
                warnings.warn(
                    'stream message not available. Returning immediately')
                pbar.close()
                return
            pbar.update(progress.incrementSec)
            if self.dsp_stream[stream_id]["stream"].done():
                pbar.update(progress.incrementSec)
                break
            time.sleep(0.1)
        pbar.close()

    def dsp_progress(self, stream_id):
        """
        Returns the progress of the 'stream_id' DSP stream running in the background.  It returns 1.0 if the DSP operation has completed.
        It is non-blocking.

        Parameters:
            stream_id (str): stream ID of the requested DSP stream.

        Returns:
            frac_complete (float): fraction complete of the DSP stream.

        Examples:
            >>> stream_id = client.dsp_low_pass_filter(750, './my_source_file', './my_output_file')
            >>> dsp_progress(stream_id)
            >>> 0.5

        See Also:
            :py:meth:`dsp_progress_bar`
            :py:meth:`dsp_low_pass_filter`

        """
        if self.dsp_stream[stream_id]["stream"].done():
            return 1.0
        for progress in self.dsp_stream[stream_id]["stream"]:
            if progress.fracComplete is None:
                warnings.warn(
                    'stream message not available. Returning immediately')
                return
            return progress.fracComplete

    def dsp_low_pass_filter(
        self,
        cutoff_freq: float,
        source_path: any,
        target_path: any,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ) -> str:
        """
        Applies low-pass filter to source file to result in target file.

        The requested cut-off frequency must be < fs/2, where fs is the sample frequency of the source file.

        Parameters:
            cutoff_freq (float): filter cut-off frequency in samples/sec (Hz)
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files must be XDAT files.
            This is non-blocking and the requested DSP operation runs in the background.

        Example:
            >>> client.dsp_low_pass_filter(750, './my_source_file', './my_output_file')
            my_stream_id_e4c-jnwq

        See Also:
            :py:meth:`dsp_progress`
            :py:meth:`dsp_progress_bar`
            :py:meth:`dsp_high_pass_filter`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('lp_filter')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode('filt_lp').lowpass(cutoff_freq))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['filt_lp'].id))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(target_path)))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['filt_lp'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {
            'data_source': self._link_data_file(
                Path(source_path)),
            'protocol': protocol,
            'stream': self._do_protocol(protocol, hub_name),
        }
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_high_pass_filter(
        self,
        cutoff_freq: float,
        source_path: str | Path,
        target_path: str | Path,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ):
        """
        Applies high-pass filter to source file to result in target file.

        The requested cut-off frequency must be < fs/2, where fs is the sample frequency of the source file.

        Parameters:
            cutoff_freq (float): filter cut-off frequency in samples/sec (Hz)
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files must be XDAT files.
            This is non-blocking and the requested DSP operation runs in the background.

        Example:
            >>> client.dsp_high_pass_filter(20, './my_source_file', './my_output_file')
            None
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('hp_filter')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode('filt_hp').highpass(cutoff_freq))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['filt_hp'].id))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(target_path)))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['filt_hp'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(Path(source_path))
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_band_pass_filter(
        self,
        low_cutoff_freq: float,
        high_cutoff_freq,
        source_path: any,
        target_path: any,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ) -> str:
        """
        Applies band-pass filter to source file to result in target file.

        The requested cut-off frequencies must be < fs/2, where fs is the sample frequency of the source file.
        The low_cutoff_freq must be less than the high_cutoff_freq.

        Parameters:
            low_cutoff_freq (float): filter low cut-off frequency in samples/sec (Hz)
            high_cutoff_freq (float): filter high cut-off frequency in samples/sec (Hz)
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files must be XDAT files.
            This is non-blocking. The requested DSP operation runs in the background.

        Example:
            >>> client.dsp_band_pass_filter(750,  2000, './my_source_file', './my_output_file')
            my_stream_id_e4c-jtyn4

        See Also:
           :py:meth:`dsp_progress`
           :py:meth:`dsp_progress_bar`
           :py:meth:`dsp_high_pass_filter`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('bp_filter')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode('filt_bp').bandpass(
            low_cutoff_freq, high_cutoff_freq))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['filt_bp'].id))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(target_path)))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['filt_bp'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_band_stop_filter(
        self,
        low_cutoff_freq: float,
        high_cutoff_freq,
        source_path: any,
        target_path: any,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ) -> str:
        """
        Applies band-stop (aka band reject) filter to source file to result in target file.

        The requested cut-off frequencies must be < fs/2, where fs is the sample frequency of the source file.
        The low_cutoff_freq must be less than the high_cutoff_freq.

        Parameters:
            low_cutoff_freq (float): filter low cut-off frequency in samples/sec (Hz)
            high_cutoff_freq (float): filter high cut-off frequency in samples/sec (Hz)
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files must be XDAT files.
            This is non-blocking. The requested DSP operation runs in the background.

        Example:
            >>> client.dsp_band_stop_filter(750,  2000, './my_source_file', './my_output_file')
            my_stream_id_e4c-jtyn4

        See Also:
           :py:meth:`dsp_progress`
           :py:meth:`dsp_progress_bar`
           :py:meth:`dsp_high_pass_filter`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('bs_filter')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode('filt_bs').bandstop(
            low_cutoff_freq, high_cutoff_freq))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['filt_bs'].id))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(target_path)))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['filt_bs'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_notch_filter(
        self,
        freq: float,
        bandwidth,
        source_path: any,
        target_path: any,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ) -> str:
        """
        Applies notch filter to source file to result in target file.

        The requested frequency must be < fs/2, where fs is the sample frequency of the source file.

        Parameters:
            freq (float): notch frequency in samples/sec (Hz)
            bandwidth (float): filter bandwidth in samples/sec (Hz)
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files must be XDAT files.
            This is non-blocking. The requested DSP operation runs in the background.

        Example:
            >>> client.dsp_notch_filter(50,  2, './my_source_file', './my_output_file')
            my_stream_id_e4c-lkqw

        See Also:
           :py:meth:`dsp_progress`
           :py:meth:`dsp_progress_bar`
           :py:meth:`dsp_high_pass_filter`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('notch_filter')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode('filt_notch').notch(freq, bandwidth))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['filt_notch'].id))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(target_path)))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['filt_notch'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_paired_diff(
        self,
        source_chan: int,
        ref_chan: int,
        source_path: any,
        target_path: any,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ) -> str:
        """
        Applies paired differential transform to the source file resulting in the target file.

        The source channel is set to the new value = (source-reference) for each time sample.
        The reference signal is zeroed.

        Parameters:
            source_chan (int): source channel index
            ref_chan (int): reference channel index
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files must be XDAT files.
            This is non-blocking. The requested DSP operation runs in the background.

        Example:
            >>> client.dsp_paired_diff(1,  5, './my_source_file', './my_output_file')
            my_stream_id_e4c-rt4m

        See Also:
           :py:meth:`dsp_progress`
           :py:meth:`dsp_progress_bar`
           :py:meth:`dsp_high_pass_filter`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('paired_diff')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode(
            'diff').paired_ref(source_chan, ref_chan))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['diff'].id))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(target_path)))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['diff'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_virtual_ref(
        self,
        ref_chan: int,
        source_path: any,
        target_path: any,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ) -> str:
        """
        Applies virtual reference transform to the source file resulting in the target file.

        All channels are set to the new value = (chan-ref_chan) for each time sample.
        The reference channel is zeroed.

        Parameters:
            ref_chan (int): reference channel index
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files must be XDAT files.
            This is non-blocking. The requested DSP operation runs in the background.

        Example:
            >>> client.dsp_virtual_ref(5, './my_source_file', './my_output_file')
            my_stream_id_e4c-wq5c

        See Also:
           :py:meth:`dsp_progress`
           :py:meth:`dsp_progress_bar`
           :py:meth:`dsp_high_pass_filter`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('virtual_ref')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode('vref').virtual_ref(ref_chan))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['vref'].id))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(target_path)))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['vref'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_car(
        self, source_path: any, target_path: any, force=False, hub_name=DEFAULT_HUB_ID
    ) -> str:
        """
        Applies common average reference transform to the source file resulting in the target file.

        All channels are set to the new value = (chan-v_ref) for each time sample, where v_ref is the sum of all the channels.

        Parameters:
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files must be XDAT files.
            This is non-blocking. The requested DSP operation runs in the background.

        Example:
            >>> client.dsp_car('./my_source_file', './my_output_file')
            my_stream_id_e4c-wq5c

        See Also:
           :py:meth:`dsp_progress`
           :py:meth:`dsp_progress_bar`
           :py:meth:`dsp_high_pass_filter`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('CAR')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode('car').car())
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['car'].id))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(target_path)))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['car'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_time_decimate(
        self,
        decimate: int,
        source_path: Path,
        target_path: Path,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ) -> str:
        """
        Applies time decimation transform to the source file resulting in the target file.

        All channels are decimated in time by 'decimate'.
        The target file sampling frequency is equal to the source file sampling frequency divided by 'decimate'
        For example, if 'decimate=2' and 'source_sampling_freq=10000', then the target file consists of every two source channel samples
        and 'target_sampling_freq=5000'

        Parameters:
            decimate (int): sample factor
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files must be XDAT files.
            This is non-blocking. The requested DSP operation runs in the background.

        Example:
            >>> client.dsp_time_decimate(3, './my_source_file', './my_output_file')
            my_stream_id_e4c-wq5c

        See Also:
           :py:meth:`dsp_progress`
           :py:meth:`dsp_progress_bar`
           :py:meth:`dsp_high_pass_filter`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('decimate')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode('dec').time_decimate(decimate))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['dec'].id))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(target_path)))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['dec'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def file_convert(
            self,
            source_path: Path,
            dest_path: Path,
            dest_file_type: str,
            force=False,
            hub_name=DEFAULT_HUB_ID,
    ) -> str:
        """
        Converts source file to target file.

        Parameters:
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            The source and target files
        """

        source_path = Path(source_path).expanduser().resolve()
        source_path = Path(source_path.parent, source_path.stem)

        dest_path = Path(dest_path).expanduser().resolve()
        dest_path = Path(dest_path.parent, dest_path.stem)

        # load source
        try:
            source_meta = self._link_data_file(Path(source_path))
        except Exception as e:
            raise ValueError("error loading source file", e)

        if dest_path.suffix != "":
            dest_file_type = RadiensFileType.parse(
                ext_to_radiens_file_type(dest_path.suffix))
        elif Path('file.'+dest_file_type).suffix != "":
            dest_file_type = RadiensFileType.parse(
                ext_to_radiens_file_type(Path('file.'+dest_file_type).suffix))
        else:
            warnings.warn(
                'no file extension provided for target file. Assuming XDAT')
            dest_file_type = RadiensFileType.XDAT

        source_file_type = source_meta.file_type
        if source_file_type == dest_file_type and dest_path.parent == source_path.parent and dest_path.stem == source_path.stem:
            raise ValueError("source and target files are the same")

        if source_file_type not in CONVERTIBLE_RADIENS_FILE_TYPES or dest_file_type not in CONVERTIBLE_RADIENS_FILE_TYPES:
            raise ValueError(
                f"file conversion between {source_file_type} and {dest_file_type} is not supported")

        source_path = Path(str(source_path)+to_suffix(source_file_type))
        dest_path = Path(str(dest_path)+to_suffix(dest_file_type))

        if force:
            try:
                self._fsys.delete_files(dest_path)
            except RpcException as e:
                if "datasource does not exist" in map(
                        lambda x: x.strip(), e.message.split(':')):
                    pass
                else:
                    raise e

        source_path = str(source_path)
        protocol = ProtocolAPI('file_convert')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(Path(source_path)))
        protocol.add_node(TransformNode(
            'targ_0').datasource_sink(Path(dest_path)))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_slice_time(
        self,
        time_start: float,
        time_end: float,
        source_path: Path,
        target_path: Path,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ):
        """
        Slices the source data file in time to result in the target file.

        Parameters:
            time_start (float): slice start time in seconds
            time_end (float): slice end time in seconds
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            "time_start" and "time_end" are the absolute time in the source file.
            The source and target files must be XDAT files.
            This is non-blocking and the requested DSP operation runs in the background.

        Example:
            >>> client.dsp_slice_time(0.5, 10, './my_source_file', './my_output_file')
            None

        See Also:
            :py:meth:`dsp_progress`
            :py:meth:`dsp_progress_bar`
            :py:meth:`dsp_slice_channels`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('slice_time_0')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(source_path))
        protocol.add_node(TransformNode(
            'slice_time').slice_time(time_start, time_end))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['slice_time'].id))
        protocol.add_node(TransformNode('targ_0').datasource_sink(target_path))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['slice_time'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(
            Path(source_path)), 'protocol': protocol, 'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def dsp_slice_channels(
        self,
        target_channels: list,
        source_path: Path,
        target_path: Path,
        force=False,
        hub_name=DEFAULT_HUB_ID,
    ):
        """
        Slices out (removes) the target channels from source data file to result in the target file.

        Parameters:
            target_channels (list): channel indices of channels to slice from the source
            source_path (str, pathlib.Path): path to source file
            target_path (str, pathlib.Path): path to target (output) file
            force (bool): flag to force replacing target file if it exists
            hub_name (str): radiens hub name (default=radiens.utils.constants.DEFAULT_HUB)

        Returns:
            stream_id (str): background DSP stream ID

        Notes:
            "target_channels" are the absolute channel indices of the source file, including GPIO signals.
            The source and target files must be XDAT files.
            This is non-blocking and the requested DSP operation runs in the background.

        Example:
            >>> client.dsp_slice_channels([0,2,8,5], './my_source_file', './my_output_file')
            None

        See Also:
           :py:meth:`dsp_progress`
           :py:meth:`dsp_progress_bar`
           :py:meth:`dsp_slice_time`
        """
        if force:
            rm_xdat_file(Path(target_path))
        protocol = ProtocolAPI('slice_chan_0')
        protocol.add_node(TransformNode(
            'src_0').datasource_source(source_path))
        protocol.add_node(TransformNode(
            'slice_chan').slice_channels(target_channels))
        protocol.add_edge(TransformEdge(
            'edge_0', protocol.nodes['src_0'].id, protocol.nodes['slice_chan'].id))
        protocol.add_node(TransformNode('targ_0').datasource_sink(target_path))
        protocol.add_edge(TransformEdge(
            'edge_1', protocol.nodes['slice_chan'].id, protocol.nodes['targ_0'].id))
        self._dsp_stream[protocol.id] = {'data_source': self._link_data_file(Path(source_path)),
                                         'protocol': protocol,
                                         'stream': self._do_protocol(protocol, hub_name)}
        self._clear_dataset(source_path)
        self._dsp_stream[protocol.id]["data_source"].clear_dataset_id()
        return protocol.id

    def _do_protocol(self, protocol: ProtocolAPI, hub_name=DEFAULT_HUB_ID) -> str:
        api_curate.set_protocol(self._server_address(
            hub_name, RadiensService.CORE), protocol)
        return api_curate.apply_protocol(self._server_address(hub_name, RadiensService.CORE),  protocol.id)

    def _clear_dataset(self, source=[], dataset_id=[], hub_name=DEFAULT_HUB_ID) -> str:
        if not isinstance(dataset_id, (list, str, Iterable)):
            raise ValueError("dataset_id must be string or list of strings")
        if isinstance(dataset_id, str):
            dataset_id = [dataset_id]
        if isinstance(source, (str, Path)):
            source = [Path(source)]
        _get_ids = []
        for _id in self._get_dataset_ids(hub_name):
            if len(dataset_id) > 0 and dataset_id[0] == "all" and _id.find(self.id) > 0:
                _get_ids.append(_id)
            else:
                for req_id in dataset_id:
                    if req_id == _id:
                        _get_ids.append(_id)
            for src in source:
                if _id.find(Path(src).stem) > 0 and _id.find(self.id) > 0:
                    _get_ids.append(_id)
        return api_videre.unlink_datasource(
            self._server_address(hub_name, RadiensService.CORE), _get_ids
        )

    def _get_dataset_ids(self, hub_name=DEFAULT_HUB_ID) -> list:
        return api_videre.list_datasource_ids(
            self._server_address(hub_name, RadiensService.CORE)
        )

    def _link_data_file(
        self, source: any, calc_metrics=False, force=False, hub_name=DEFAULT_HUB_ID
    ) -> DatasetMetadata:
        if isinstance(source, (Path, str)):
            source = Path(source).expanduser().resolve()
        else:
            raise ValueError("source must be string or Path")
        req = datasource_pb2.DataSourceSetSaveRequest(
            path=str(source.parent),
            baseName=source.stem,
            fileType=to_radiens_file_type(source).value,
            dsourceID=source.stem + self.id,
            isBackgroundKPI=calc_metrics,
            isForce=force,
        )
        return api_videre.set_datasource(
            self._server_address(hub_name, RadiensService.CORE), req
        )



import time
import unittest
from pathlib import Path
from pprint import pprint

from radiens.utils.enums import SignalType, TrsMode
from radiens.utils.util import make_time_range
from radiens.videre_client import VidereClient


class base_utest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._dir = Path(__file__).parent

    def setUp(self):
        self.client = VidereClient()
        self.radix_data = '~/radix/data'
        self.data_path = '~/radix/data/radiens_test_data'
        self.dataset_0 = '64chan_1s'
        self.tmp_data_path = '~/radix/data/radiens_test_data/tmp'
        self.tmp_dataset_1 = 'tmp_0'

    def tearDown(self) -> None:
        self.client.clear_dataset(dataset_id='all')

    def test_get_signals(self):
        dsource = self.client.link_data_file(
            Path(self.data_path, '64chan_1s.xdat'))
        chunk = 0.01
        TR = make_time_range(time_range=[
                             dsource.time_range.sec[0], dsource.time_range.sec[0] + chunk], fs=dsource.time_range.fs)
        sig = self.client.signals().get_signals(
            TR, TrsMode.SUBSET, sig_sel=None, dataset_metadata=dsource)
        self.assertEqual(
            sig.signals.amp.shape[0], dsource.channel_metadata.num_sigs(SignalType.AMP))
        self.assertEqual(
            sig.signals.gpio_ain.shape[0], dsource.channel_metadata.num_sigs(SignalType.AIN))
        self.assertEqual(2, sig.signals.gpio_ain.shape[0])
        self.assertEqual(sig.signals.amp.shape[1], TR.N)
        self.assertEqual(sig.signals.amp.shape[1], sig.time_range.N)
        self.assertEqual(sig.signals.amp.shape[1], TR.fs * chunk)
        self.assertEqual(dsource.time_range.sec[0], sig.time_range.sec[0])
        self.assertEqual(
            dsource.time_range.sec[0]+chunk, sig.time_range.sec[1])
        self.assertEqual(dsource.time_range.fs, sig.time_range.fs)

    def test_get_psd(self):
        dsource = self.client.link_data_file(
            Path(self.data_path, '64chan_1s.xdat'))
        chunk = 0.2
        TR = make_time_range(time_range=[
                             dsource.time_range.sec[0], dsource.time_range.sec[0] + chunk], fs=dsource.time_range.fs)
        psd = self.client.signals().get_psd(TR, TrsMode.SUBSET, sig_sel=None,
                                            samp_freq=2000.0, dataset_metadata=dsource)
        self.assertEqual(
            psd.psd.shape[0], dsource.channel_metadata.num_sigs(SignalType.AMP))

    def test_signal_metrics(self):
        if Path(self.data_path, '64chan_1s.kpi').expanduser().absolute().exists():
            Path(self.data_path, '64chan_1s.kpi').expanduser().absolute().unlink()
        self.assertFalse(
            Path(self.data_path, '64chan_1s.kpi').expanduser().absolute().exists())

        dsource = self.client.link_data_file(
            Path(self.data_path, '64chan_1s.xdat'), calc_metrics=False)
        self.assertEqual("64chan", dsource.id[0:6])
        time.sleep(2)
        self.assertFalse(
            Path(self.data_path, '64chan_1s.kpi').expanduser().absolute().exists())

        status = self.client.signal_metrics().get_metrics_status(dsource)
        self.assertEqual([0, 0], status.time_range.sec.tolist())

        self.client.signal_metrics().calculate(dsource)
        time.sleep(2)
        self.assertTrue(
            Path(self.data_path, '64chan_1s.kpi').expanduser().absolute().exists())
        status = self.client.signal_metrics().get_metrics_status(dsource)
        self.assertEqual([16, 17], status.time_range.sec.tolist())

        self.client.signal_metrics().clear(dsource)
        self.assertFalse(
            Path(self.data_path, '64chan_1s.kpi').expanduser().absolute().exists())

        self.client.clear_dataset(dataset_id='all')
        dsource = self.client.link_data_file(
            Path(self.data_path, '64chan_1s.xdat'), calc_metrics=True)
        self.assertEqual("64chan", dsource.id[0:6])
        time.sleep(2)
        self.assertTrue(
            Path(self.data_path, '64chan_1s.kpi').expanduser().absolute().exists())


if __name__ == "__main__":
    unittest.main()



import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from radiens.api import api_allego, api_videre
from radiens.grpc_radiens import biointerface_pb2, radiens_dev_pb2
from radiens.lib.channel_metadata import ChannelMetadata
from radiens.utils.constants import (DEFAULT_HUB_ID, NEURONS1_ADDR,
                                     NEURONS_SINK_DSOURCE_ID, TimeRange)
from radiens.utils.enums import RadiensService, SignalType
from radiens.utils.util import make_time_range

# the following lines are to avoid circular imports and are only used for typing hints
# (TYPE_CHECKING always evaluates to false at runtime)
if TYPE_CHECKING:
    from radiens.allego_client import AllegoClient
    from radiens.videre_client import VidereClient


class SpikesClient():
    """
    Spikes object for AllegoClient, VidereClient
    """

    def __init__(self, parent_client):
        """
        """
        self.__parent: VidereClient | AllegoClient = parent_client

    def _is_allego(self) -> bool:
        return self.__parent.type.is_allego()

    def _is_curate(self) -> bool:
        return self.__parent.type.is_curate()

    def _is_videre(self) -> bool:
        return self.__parent.type.is_videre()

    def get_spikes_metadata(self):
        """
        Returns metadata on the spikes processed by the spike sorter

        Returns:
            pandas.DataFrame

        Example:
            >>> client.get_spikes()
            None

        """
        if self._is_allego():
            return api_allego.get_spikes_spec(NEURONS1_ADDR)

    def get_spike_rasters(self, dsource_id=None):
        """
        Returns spike rasters from the spike sorter

        Returns:
            pandas.DataFrame

        Example:
            >>> client.get_spikes()
            None

        """
        msg = biointerface_pb2.SpikeSorterGetRasterDataRequest()
        if self._is_allego():
            return api_allego.sorter_get_raster_data(NEURONS1_ADDR, msg)

    def get_recent_spikes(self, time_wdw_sec=1.0, ntv_chan_idxs=None, max_spikes_per_chan=100, channel_meta=None):
        """
        Returns the most recent spikes for the requested ``ntv_chan_idxs`` over the requested time window. 
        The returned spike data has spike timestamps, spike waveforms, and spike labels. 

        Returns:
            pandas.DataFrame

        Example:
            >>> client.get_spikes()
            None

        """
        if channel_meta in [None]:
            channel_meta = self.__parent.get_channel_metadata()
        if not isinstance(channel_meta, ChannelMetadata):
            raise ValueError(
                'channel metadata must be of type ChannelMetadata')
        ntv_chan_idxs = channel_meta.index(SignalType.AMP).ntv
        msg = biointerface_pb2.SpikesGetSpikesRequest(dsourceID=NEURONS_SINK_DSOURCE_ID,
                                                      mode=biointerface_pb2.Spikes,  # not used
                                                      ntvChanIdx=ntv_chan_idxs,
                                                      timeStart=time_wdw_sec,
                                                      timeStop=np.nan,
                                                      spikeLabels=[],  # not used
                                                      maxSpikesPerChannel=max_spikes_per_chan,
                                                      spikeComp=[],  # not used
                                                      )
        return api_allego.get_spikes(NEURONS1_ADDR, msg)

    def get_spike_waveforms(self, dataset_id: str, time_range: TimeRange | list = None, hub_name=DEFAULT_HUB_ID):
        """
        Gets spike waveforms for specified time range of a spikes file

        Parameters: 
            dataset_id (str): dataset ID 
            time_range (list[float])
            ntv_idxs (list[float])

        Returns:

        """

        if self._is_videre():
            time_range = self.__parent._parse_time_range(
                dataset_id, time_range)
            return api_videre.get_spike_waveforms(self.__parent._server_address(hub_name, RadiensService.CORE), dataset_id, time_range)
        else:
            raise Exception('{} cannot get spike waveforms',
                            self.__parent.type)

    def get_spike_timestamps(self, dataset_id: str, time_range:  TimeRange | list = None, hub_name=DEFAULT_HUB_ID):
        """
        Gets spike timestamps for specified time range of a spikes file

        Parameters: 
            dataset_id (str): dataset ID 
            time_range (list[float])

        Returns:

        """

        if self._is_videre():
            time_range = self.__parent._parse_time_range(
                dataset_id, time_range)
            return api_videre.get_spikes_timestamps(
                self.__parent._server_address(hub_name, RadiensService.CORE),
                dataset_id,
                time_range,
            )
        else:
            raise Exception('{} cannot get spike timestamps',
                            self.__parent.type)

    def get_neuron_timestamps(self, dataset_id: str, time_range: TimeRange, ntv_idxs: list):
        resp = self.get_spike_timestamps(dataset_id, time_range, ntv_idxs)
        neuron_dict = {}
        for chan in resp:
            for i, l in enumerate(chan['labels']):
                unique_label = str(chan['ntv_idx']) + str(l)
                if unique_label not in neuron_dict:
                    neuron_dict[unique_label] = {
                        'timestamps': [],
                        'ntv_idx': chan['ntv_idx']
                    }
                neuron_dict[unique_label]['timestamps'].append(
                    chan['timestamps'][i])

        # rename neurons to be numbers 1-len(neuron_dict)
        neuron_labels_old = list(neuron_dict.keys())
        for i, unique_label in enumerate(neuron_labels_old):
            neuron_dict[i] = neuron_dict.pop(unique_label)
        return neuron_dict

    def get_neurons(self):
        """
        Returns neurons from the spike sorter

        Returns:
            pandas.DataFrame

        Example:
            >>> client.get_spikes()
            None

        """
        msg = biointerface_pb2.BiointerfaceGetNeuronsRequest()
        return api_allego.get_neurons(NEURONS1_ADDR, msg)

    def convert_kilosort_output(self,
                                source_bio_path: str | Path = "",
                                source_bio_basename: str | Path = "",
                                source_vex_path: str | Path = "",
                                source_vex_basename: str | Path = "",
                                sink_bio_path: str | Path = "",
                                sink_bio_basename: str | Path = "",
                                hub_name=DEFAULT_HUB_ID):
        """
         Converts the output of kilosort into a .spikes file that can be loaded into radiens
        """
        spec = radiens_dev_pb2.WrangleMerge(
            vexSpikesView=radiens_dev_pb2.WrangleMergeVexSpikesView(
                sourceVex=radiens_dev_pb2.WrangleFileDesc(
                    path=str(source_vex_path),
                    baseName=str(source_vex_basename),
                ),
                sourceBio=radiens_dev_pb2.WrangleFileDesc(
                    path=str(source_bio_path),
                    baseName=str(source_bio_basename),
                ),
                view=[
                    radiens_dev_pb2.WrangleMergeVexSpikesViewSpec(
                        sinkBio=radiens_dev_pb2.WrangleFileDesc(
                            path=str(sink_bio_path),
                            baseName=str(sink_bio_basename),
                        )
                    )
                ],
            ),
        )

        req = radiens_dev_pb2.WrangleRequest(
            mode=radiens_dev_pb2.WRANGLE_MODE_IMPORT_KILOSORT_SPIKES,
            spec=spec,
        )

        # radiens_dev_pb2.WrangleMode.DESCRIPTOR.
        if not self._is_videre():
            raise Exception(
                '{} cannot convert kilosort output'.format(self.__parent.type))
        return api_videre.convert_kilosort_output(self.__parent._server_address(hub_name, RadiensService.SORTER), req)



"""
radiens CLI
"""
from pathlib import Path

import click
import pandas as pd
from radiens.cli.signal_metrics import metrics
from radiens.cli.spike_sorter import sorter
from radiens.utils.config import ERROR_COLOR_FG
from radiens.utils.util import time_now


@click.group(short_help="allego commands")
@click.pass_context
def allego(ctx):
    """
    ALLEGO command group

    This command group works concurrently with the ALLEGO app to control the data acquisition system.
    """
    ctx.obj["cmd_app"] = "allego"


allego.add_command(sorter)
allego.add_command(metrics)


@allego.command(short_help="control streaming or print the system status")
@click.pass_context
@click.argument("arg", type=click.STRING, nargs=-1)
@click.option(
    "-v",
    "--verbose",
    type=click.Choice(["brief", "all"], case_sensitive=False),
    default="brief",
    help="detail of status",
    show_default=True,
)
def stream(ctx, arg, verbose):
    """
    Turns streaming on or off or prints the stream status.

    arg: [[on | off]], no argument will print status
    """
    if len(arg) == 0:
        status(ctx, verbose)
        return
    if arg[0] not in ["on", "off"]:
        click.echo(click.style(
            "`arg` must be `on` or `off`", fg=ERROR_COLOR_FG))
        return
    try:
        ctx.obj["allego_client"].set_streaming(arg[0])
    except Exception as ex:
        click.echo("Error: {}".format(ex))
        return
    click.echo("ALLEGO stream set to {} at {}".format(arg[0], time_now()))


@allego.command(short_help="prints system status")
@click.pass_context
@click.option(
    "-v",
    "--verbose",
    type=click.Choice(["brief", "all"], case_sensitive=False),
    default="brief",
    help="detail of status",
    show_default=True,
)
def status(ctx, verbose):
    """
    Copy one Radiens-aware data file set to a 'dest' data file set.
    """
    try:
        resp = ctx.obj["allego_client"].get_status()
    except Exception as ex:
        click.echo("Error: {}".format(ex))
        return
    resp.print(verbose)


@allego.command(short_help="control recording or print system status")
@click.pass_context
@click.argument("arg", type=click.STRING, nargs=-1)
@click.option(
    "-v",
    "--verbose",
    type=click.Choice(["brief", "all"], case_sensitive=False),
    default="brief",
    help="detail of status",
    show_default=True,
)
def record(ctx, arg, verbose):
    """
    Turns recording on or off or prints the stream status.

    arg: [[on | off]], no argument will print status

    Notes:
        1. streaming is always turned on when recording is started.
    """
    if len(arg) == 0:
        status(ctx, verbose)
        return
    if arg[0] not in ["on", "off"]:
        click.echo(click.style(
            "`arg` must be `on` or `off`", fg=ERROR_COLOR_FG))
        return
    try:
        ctx.obj["allego_client"].set_recording(arg[0])
    except Exception as ex:
        click.echo("Error: {}".format(ex))
        return
    stat = ctx.obj["allego_client"].get_status()
    if arg[0] in ["on"]:
        click.echo(
            "ALLEGO recording {} ({}); Saving to {}".format(
                arg[0], time_now(), Path(stat.recording.path,
                                         stat.recording.file_name)
            )
        )
    else:
        click.echo(
            "ALLEGO recording {} ({}); Recorded {:.3f} sec to {}".format(
                arg[0],
                time_now(),
                stat.recording.dur_sec,
                Path(stat.recording.path, stat.recording.file_name),
            )
        )


@allego.command(short_help="restart allego with hardware or simulator")
@click.pass_context
@click.option(
    "-sys",
    "--system",
    type=click.Choice(
        [
            "sbpro",
            "sbpro-sinaps-256",
            "sbpro-sinaps-1024" "sbclassic",
            "sim-sine",
            "sim-spikes",
            "open-ephys_usb2",
            "open-ephys_usb3",
            "intan1024",
            "intan512",
            "xdaq-one-rec",
            "xdaq-one-stim",
            "xdaq-core-rec",
            "xdaq-core-stim",
        ],
        case_sensitive=False,
    ),
    default="sim-spikes",
    help="allego system type",
    show_default=True,
)
def restart(ctx, system):
    """
    Restart Allego for specified hardware or simulator system.
    """
    try:
        ctx.obj["allego_client"].restart(system)
    except Exception as ex:
        click.echo("Error: {}".format(ex))
        return
    stat = ctx.obj["allego_client"].get_status()
    print("allego restarted using {}".format(stat.system_mode))


@allego.group(short_help="set allego settings")
@click.pass_context
def set(ctx):
    """
    Set Allego settings

    This command group has sub-commands to set various types of system settings
    """


@set.command("system", short_help="set basic system settings")
@click.pass_context
@click.option(
    "-fs",
    "--samp_freq",
    type=click.INT,
    default=-1,
    help="system sample frequency [samples/sec]",
    show_default=True,
)
def set_system(ctx, samp_freq):
    """
    Sets the basic system settings.

    """
    try:
        if samp_freq > 0:
            ctx.obj["allego_client"].set_sampling_freq(float(samp_freq))
    except Exception as ex:
        click.echo("Error: {}".format(ex))
        return
    stat = ctx.obj["allego_client"].get_status()
    if samp_freq > 0:
        print("new sampling freq = {}".format(stat.sample_freq))


@set.command("record", short_help="set recording parameters")
@click.pass_context
@click.option(
    "-p", "--path", type=click.STRING, default="", help="path", show_default=True
)
@click.option(
    "-n",
    "--name",
    type=click.STRING,
    default="",
    help="base name for composed file name",
    show_default=True,
)
@click.option(
    "-i",
    "--index",
    type=click.INT,
    default=-1,
    help="index for composed file name",
    show_default=True,
)
@click.option(
    "-t",
    "--timestamp",
    type=click.Choice(["-1", "0", "1"]),
    default="-1",
    help="timestamp for file name",
    show_default=True,
)
def set_record(ctx, path, name, index, timestamp):
    """
    Sets the recording settings.

    \b
    --path option value:
        any: path for saving the recording file
        '' : no change
    \b
    --timestamp option value:
        0:  do not include timestamp in composed file name
        1:  do include timestamp in composed file name
        -1: no change

    \b
    --index option value:
        idx >= 0:  index value to include in composed file name
        -1: no change
    """
    try:
        stat = ctx.obj["allego_client"].get_status()
        path = (
            str(Path(path).expanduser().absolute().parent)
            if len(path) > 0
            else stat.recording.path
        )
        name = name if len(name) > 0 else stat.recording.base_name
        index = index if index >= 0 else stat.recording.index
        if timestamp in ["0"]:
            _ts = False
        elif timestamp in ["1"]:
            _ts = True
        else:
            _ts = False if len(stat.recording.timestamp) == 0 else True
        ctx.obj["allego_client"].set_recording_config(path, name, index, _ts)

    except Exception as ex:
        click.echo("Error: {}".format(ex))
        return
    stat = ctx.obj["allego_client"].get_status()
    click.echo("new recording file {}".format(stat.recording.file_name))


@set.command("record", short_help="set recording parameters")
@click.pass_context
@click.option(
    "-p", "--path", type=click.STRING, default="", help="path", show_default=True
)
@click.option(
    "-n",
    "--name",
    type=click.STRING,
    default="",
    help="base name for composed file name",
    show_default=True,
)
@click.option(
    "-i",
    "--index",
    type=click.INT,
    default=-1,
    help="index for composed file name",
    show_default=True,
)
@click.option(
    "-t",
    "--timestamp",
    type=click.Choice(["-1", "0", "1"]),
    default="-1",
    help="timestamp for file name",
    show_default=True,
)
def set_record_params(ctx, path, name, index, timestamp):
    """
    Sets the recording settings.

    \b
    --path option value:
        any: path for saving the recording file
        '' : no change
    \b
    --timestamp option value:
        0:  do not include timestamp in composed file name
        1:  do include timestamp in composed file name
        -1: no change

    \b
    --index option value:
        idx >= 0:  index value to include in composed file name
        -1: no change
    """
    try:
        stat = ctx.obj["allego_client"].get_status()
        path = (
            str(Path(path).expanduser().absolute().parent)
            if len(path) > 0
            else stat.recording.path
        )
        name = name if len(name) > 0 else stat.recording.base_name
        index = index if index >= 0 else stat.recording.index
        if timestamp in ["0"]:
            _ts = False
        elif timestamp in ["1"]:
            _ts = True
        else:
            _ts = False if len(stat.recording.timestamp) == 0 else True
        ctx.obj["allego_client"].set_recording_config(path, name, index, _ts)

    except Exception as ex:
        click.echo("Error: {}".format(ex))
        return
    stat = ctx.obj["allego_client"].get_status()
    click.echo("new recording file {}".format(stat.recording.file_name))



from __future__ import annotations

import numbers
from collections import namedtuple
from enum import Enum, auto
from typing import Any, Type, TypeVar

from radiens.grpc_radiens import allegoserver_pb2, common_pb2, spikesorter_pb2

MetricID = namedtuple("METRIC_ID", ['mode', 'name'])
"""
Example:
    >>> metric_id = MetricID(mode=MetricMode.BASE, name=MetricName.MEAN)
"""


MetricStats = namedtuple("METRIC_STATS", ['mean', 'sd', 'max', 'min'])

# enums


class ClientType(Enum):
    """TYPE is used to specify the type of client. 
    """

    ALLEGO = 'AllegoClient'  #: Allego client
    VIDERE = 'VidereClient'  #: Videre client
    CURATE = 'CurateClient'  #: Curate client

    @classmethod
    def parse(cls, x: str | ClientType) -> ClientType:
        if isinstance(x, str):
            if x.lower() == 'allego' or x.lower() == 'allegoclient':
                return cls.ALLEGO
            elif x.lower() == 'videre' or x.lower() == 'videreclient':
                return cls.VIDERE
            elif x.lower() == 'curate' or x.lower() == 'curateclient':
                return cls.CURATE
            else:
                raise ValueError(f"Invalid {cls.__name__} value: {x}")
        elif isinstance(x, cls):
            return x
        else:
            raise TypeError(
                f"Expected str or {cls.__name__} instance, got {x.__class__}.")

    def is_allego(self):
        return self.value == ClientType.ALLEGO.value

    def is_videre(self):
        return self.value == ClientType.VIDERE.value

    def is_curate(self):
        return self.value == ClientType.CURATE.value


class RadiensService(Enum):
    """
    An enumeration of Radiens services
    """

    CORE = auto()
    DASH = auto()
    SORTER = auto()
    DEV = auto()
    RADIENSPY = auto()


class KeyIndex(Enum):
    """
    An enumeration of key index types.

    See Also:
        :py:meth:`~radiens.lib.channel_metadata.ChannelMetadata.index`

    Example:
        >>> key_ind = KeyIndex.NTV

    Valid values are as follows:
    """

    #: The *native key index* is the position of the signal (channel) trace when the backing dataset was created for each signal type. For recording files, it is the position of the channels in the DAQ data stream or primary data file. The ``ntv`` index starts at zero for each signal type.
    NTV = 0
    #: The *dataset key index* is the position of a signal trace in the backing dataset for each signal type. Unless the signal traces in the dataset were re-ordered, ``dset`` will equal ``ntv``. The ``dset`` index starts at zero for each signal type.
    DATA = 1
    #: The *system key index* starts at zero and increases in natural order over all signal types, always in the following order:
    #: ``'amp'``, ``'gpio_ain'``, ``'gpio_din'``, ``'gpio_dout'``.
    SYS = 2


# protobuf enums
T = TypeVar('T', bound='GrpcEnum')


class GrpcEnum(Enum):
    """
    A base class for all enumerations used in the gRPC API
    """
    @classmethod
    def all(cls: Type[T]) -> list[T]:
        """
        Returns a list of all enum values

        Returns:
            list[T]: A list of all enum values

        """
        return list(cls)

    @classmethod
    def all_string(cls: Type[T]) -> list[str]:
        """
        Returns a list of string names of the enum values

        Returns:
            list[str]: A list of string names of the enum values

        """
        return [e.name for e in cls]

    @classmethod
    def all_enum(cls: Type[T]) -> list[int]:
        """
        Returns a list of protobuf enum values

        Returns:
            list[int]: A list of protobuf enum values

        """
        return [e.pb_value for e in cls]

    @classmethod
    def _from_enum(cls: Type[T], pb: int) -> T:
        """
        Converts a protobuf enum value to an enum value

        Args:
            pb (int): The protobuf enum value

        Returns:
            T: The enum value

        Raises:
            ValueError: If the protobuf enum value is invalid

        """
        try:
            return cls(pb)
        except ValueError:
            raise ValueError(f"Invalid {cls.__name__} value: {pb}")

    @classmethod
    def _from_string(cls: Type[T], s: str) -> T:
        """
        Converts a string to an enum value

        Args:
            s (str): The string to convert to an enum value (case-insensitive)

        Returns:
            T: The enum value

        Raises:
            ValueError: If the string is invalid
        """
        try:
            return cls[s.upper()]
        except KeyError:
            raise ValueError(f"Invalid {cls.__name__} value: {s}")

    @classmethod
    def parse(cls: Type[T], x: str | int | T) -> T:
        """
        Converts any value to an enum value

        Args:
            x (Any): The value to convert to an enum value

        Returns:
            T: The enum value

        Raises:
            ValueError: If the value is invalid
            TypeError: If the value is not an int, str, or enum instance

        """
        if isinstance(x, numbers.Number):  # includes bool as well
            return cls._from_enum(x)
        elif isinstance(x, str):
            return cls._from_string(x)
        elif isinstance(x, cls):
            return x
        else:
            raise TypeError(
                f"Expected int, str, or {cls.__name__} instance, got {x.__class__}.")

    @classmethod
    def map_into_keys(cls: Type[T], d: dict[T | str | int]) -> dict[T]:
        """
        Parses each key and returns a new dictionary with the enums (instances) as keys. If they keys are already enum instances, they are left unchanged.
        """
        return {cls.parse(k): v for k, v in d.items()}

    @classmethod
    def map_into_keys_as_names(cls: Type[T], d: dict[T | str | int]) -> dict[str]:
        """
        Parses each key and returns a new dictionary with the enum names (i.e., strings) as keys. If the keys are already strings, they are left unchanged.
        """
        return {cls.parse(k).name: v for k, v in d.items()}

    @classmethod
    def map_into_keys_as_values(cls: Type[T], d: dict[T | str | int]) -> dict[int]:
        """
       Parses each key and returns a new dictionary with the enum values (i.e., ints) as keys. If the keys are already ints, they are left unchanged.
        """
        return {cls.parse(k).value: v for k, v in d.items()}

    @property
    def pb_value(self) -> int:
        """
        Returns the protobuf enum value
        """
        return self.value


class RasterMode(GrpcEnum):
    """
    An enumeration of raster modes
    """
    CHANNELS = spikesorter_pb2.RasterMode.CHANNELS
    NEURONS = spikesorter_pb2.RasterMode.NEURONS


class StreamMode(GrpcEnum):
    """
    An enumeration of stream modes
    """
    OFF = allegoserver_pb2.StreamMode.S_OFF
    ON = allegoserver_pb2.StreamMode.S_ON


class RecordMode(GrpcEnum):
    """
    An enumeration of record modes
    """
    OFF = allegoserver_pb2.RecordMode.R_OFF
    ON = allegoserver_pb2.RecordMode.R_ON


class DioMode(GrpcEnum):
    MANUAL = allegoserver_pb2.DIOMode.manual
    EVENTS = allegoserver_pb2.DIOMode.events
    GATED = allegoserver_pb2.DIOMode.gated


class WorkspaceApp(GrpcEnum):
    """
    An enumeration of workspace applications
    """

    ALLEGO = common_pb2.WorkspaceApp.Allego
    CURATE = common_pb2.WorkspaceApp.Curate
    VIDERE = common_pb2.WorkspaceApp.Videre


class PsdScaling(GrpcEnum):
    '''PsdScaling is used to set the power spectral density (PSD) scale. 
    '''
    ABSOLUTE = common_pb2.PSDScaling.PSD_ABSOLUTE  #: absolute scale


class SignalUnits(GrpcEnum):  # not defined in protos, but rather in corelib
    UNKNOWN = 0  #: unknown units
    MICROVOLTS = 1  #: microvolts
    VOLTS = 2  #: volts
    BINARY = 3  #: binary

    def __str__(self):
        if self.value == SignalUnits.MICROVOLTS.value:
            return "μV"
        elif self.value == SignalUnits.VOLTS.value:
            return "V"
        elif self.value == SignalUnits.BINARY.value:
            return "Bin"
        elif self.value == SignalUnits.UNKNOWN.value:
            return "Unknown"
        else:
            raise ValueError(
                f"Invalid {self.__class__.__name__} value: {self.value}")


class FftWindow(GrpcEnum):
    '''FftWindow is used to set the time-domain window function for power spectral density (PSD) analysis.  
    '''
    HAMMING_p01 = common_pb2.PSDWindowType.HAMMING_p01  #: Hamming window with 1% overlap
    #: Hamming window with 5% overlap
    HAMMING_p05 = common_pb2.PSDWindowType.HAMMING_p05
    #: pass-through (no windowing)
    PASS_THROUGH = common_pb2.PSDWindowType.PASS_THROUGH


class StatsVector(GrpcEnum):
    '''Enum for summary statistic vectors'''

    STAT_MEAN = common_pb2.SummaryStatsEnum.SS_STAT_MEAN
    STAT_SD = common_pb2.SummaryStatsEnum.SS_STAT_SD
    STAT_MODE = common_pb2.SummaryStatsEnum.SS_STAT_MODE
    STAT_MIN = common_pb2.SummaryStatsEnum.SS_STAT_MIN
    STAT_MAX = common_pb2.SummaryStatsEnum.SS_STAT_MAX
    STAT_MODE_COUNT = common_pb2.SummaryStatsEnum.SS_STAT_MODE_COUNT
    STAT_MEDIAN = common_pb2.SummaryStatsEnum.SS_STAT_MEDIAN
    STAT_Q25 = common_pb2.SummaryStatsEnum.SS_STAT_Q25
    STAT_Q75 = common_pb2.SummaryStatsEnum.SS_STAT_Q75
    STAT_SKEW = common_pb2.SummaryStatsEnum.SS_STAT_SKEW
    STAT_KURTOSIS = common_pb2.SummaryStatsEnum.SS_STAT_KURTOSIS
    STAT_N = common_pb2.SummaryStatsEnum.SS_STAT_N


class SummaryStats(GrpcEnum):
    '''Enum for spikes (aka 'neural interface') summary statistics'''

    MIN = common_pb2.SummaryStatsEnum.SS_STAT_MIN
    MAX = common_pb2.SummaryStatsEnum.SS_STAT_MAX
    MEAN = common_pb2.SummaryStatsEnum.SS_STAT_MEAN
    SD = common_pb2.SummaryStatsEnum.SS_STAT_SD
    MODE = common_pb2.SummaryStatsEnum.SS_STAT_MODE
    COUNT = common_pb2.SummaryStatsEnum.SS_STAT_MODE_COUNT
    MEDIAN = common_pb2.SummaryStatsEnum.SS_STAT_MEDIAN
    Q25 = common_pb2.SummaryStatsEnum.SS_STAT_Q25
    Q75 = common_pb2.SummaryStatsEnum.SS_STAT_Q75
    SKEW = common_pb2.SummaryStatsEnum.SS_STAT_SKEW
    KURTOSIS = common_pb2.SummaryStatsEnum.SS_STAT_KURTOSIS
    N = common_pb2.SummaryStatsEnum.SS_STAT_N


class RadiensFileType(GrpcEnum):
    """
    An enumeration of Radiens file types
    """
    RHD = common_pb2.RHD  #: RHD file
    XDAT = common_pb2.XDAT  #: XDAT file
    CSV = common_pb2.CSV  #: CSV file
    HDF5 = common_pb2.HDF5  #: HDF5 file
    NEX5 = common_pb2.NEX5  #: NEX5 file
    NWB = common_pb2.NWB  #: NWB file
    KILOSORT2 = common_pb2.KILOSORT2  #: Kilosort2 (bin) file
    NSX = common_pb2.NSX  #: NSX file
    TDT = common_pb2.TDT  #: TDT file
    SPIKES = common_pb2.SPIKES  #: Spikes file


class TrsMode(GrpcEnum):
    """
    TrsMode is the time range selector (TRS) enum used to control time range selection in calls to get signals, spikes, power spectral density, or similar data sets.  

    Example:
        >>> sel_mode = TrsMode.SUBSET 

    Given a time range [start, end) in seconds (dataset time), the selector modes are:
    """

    SUBSET = common_pb2.TimeRangeSelMode.TRS_SUBSET  #: selects [start, end)
    #: selects [start, head of stream/cache/file]
    TO_HEAD = common_pb2.TimeRangeSelMode.TRS_TO_HEAD
    #: selects [(head of stream/cache/file - start)
    FROM_HEAD = common_pb2.TimeRangeSelMode.TRS_FROM_HEAD


class MetricMode(GrpcEnum):
    BASE = common_pb2.KpiMetricsMode.Kpi_MetricsMode_Base  #: base packet
    #: running average of all base packets in the stream
    AVG = common_pb2.KpiMetricsMode.Kpi_MetricsMode_Avg
    #: cumulative over all base packets in the stream
    STREAM = common_pb2.KpiMetricsMode.Kpi_MetricsMode_Stream


class MetricName(GrpcEnum):
    #: Number of points in packet
    NUM_PTS = common_pb2.KpiMetricsEnum.KPI_NUM_PTS
    #: Number of interspike intervals in packet
    NUM_PTS_ISI = common_pb2.KpiMetricsEnum.KPI_NUM_PTS_ISI
    #: Packet duration in seconds
    DUR_SEC = common_pb2.KpiMetricsEnum.KPI_DUR_SEC
    #: Mean signal level in packet
    MEAN = common_pb2.KpiMetricsEnum.KPI_MEAN
    #: Minimum signal level in packet
    MIN = common_pb2.KpiMetricsEnum.KPI_MIN
    #: Minimum interspike signal level in packet
    MIN_ISI = common_pb2.KpiMetricsEnum.KPI_MIN_ISI
    #: Maximum signal level in packet
    MAX = common_pb2.KpiMetricsEnum.KPI_MAX
    #: Maximum interspike signal level in packet
    MAX_ISI = common_pb2.KpiMetricsEnum.KPI_MAX_ISI
    #: Absolute maximum signal level in packet
    MAX_ABS = common_pb2.KpiMetricsEnum.KPI_MAX_ABS
    #: Absolute maximum interspike signal level in packet
    MAX_ABS_ISI = common_pb2.KpiMetricsEnum.KPI_MAX_ABS_ISI
    #: Timestamp of minimum signal level in packet
    TIMESTAMP_MIN = common_pb2.KpiMetricsEnum.KPI_TIMESTAMP_MIN
    #: Timestamp of maximum signal level in packet
    TIMESTAMP_MAX = common_pb2.KpiMetricsEnum.KPI_TIMESTAMP_MAX
    #: Absolute difference between maximum and minimum signal levels in packet
    MAX_MIN_DIFF_ABS = common_pb2.KpiMetricsEnum.KPI_MAX_MIN_DIFF_ABS
    #: Absolute difference between maximum and minimum interspike signal levels in packet
    MAX_MIN_DIFF_ABS_ISI = common_pb2.KpiMetricsEnum.KPI_MAX_MIN_DIFF_ABS_ISI
    #: Signal standard deviation in packet
    SD = common_pb2.KpiMetricsEnum.KPI_SD
    #: Signal standard deviation calculated over interspike intervals
    SD_ISI = common_pb2.KpiMetricsEnum.KPI_SD_ISI
    #: Signal variance in packet
    VAR = common_pb2.KpiMetricsEnum.KPI_VAR
    #: Signal variance calculated over interspike intervals
    VAR_ISI = common_pb2.KpiMetricsEnum.KPI_VAR_ISI
    #: Root mean square (RMS) signal value in packet
    RMS = common_pb2.KpiMetricsEnum.KPI_RMS
    #: RMS signal value calculated over interspike intervals
    RMS_ISI = common_pb2.KpiMetricsEnum.KPI_RMS_ISI
    #: Noise level in microvolts
    NOISE_UV = common_pb2.KpiMetricsEnum.KPI_NOISE_UV
    #: Signal-to-noise ratio (SNR) of events in packet
    SNR = common_pb2.KpiMetricsEnum.KPI_SNR
    #: Number of events in packet
    NUM_EVENTS = common_pb2.KpiMetricsEnum.KPI_NUM_EVENTS
    #: Event rate in packet
    EVENT_RATE = common_pb2.KpiMetricsEnum.KPI_EVENT_RATE
    #: Maximum amplitude of events in packet
    EVENT_MAX = common_pb2.KpiMetricsEnum.KPI_EVENT_MAX
    #: Minimum amplitude of events in packet
    EVENT_MIN = common_pb2.KpiMetricsEnum.KPI_EVENT_MIN
    #: Absolute maximum amplitude of events in packet
    EVENT_MAX_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_MAX_ABS
    #: Absolute difference between maximum and minimum amplitudes of events in packet
    EVENT_MAX_MIN_DIFF_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_MAX_MIN_DIFF_ABS
    #: Timestamp of minimum amplitude of events in packet
    EVENT_TIMESTAMP_MIN = common_pb2.KpiMetricsEnum.KPI_EVENT_TIMESTAMP_MIN
    #: Timestamp of maximum amplitude of events in packet
    EVENT_TIMESTAMP_MAX = common_pb2.KpiMetricsEnum.KPI_EVENT_TIMESTAMP_MAX
    #: Timestamp of maximum absolute amplitude of events in packet
    EVENT_TIMESTAMP_MAX_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_TIMESTAMP_MAX_ABS
    #: Timestamp of absolute difference between maximum and minimum amplitudes of events in packet
    EVENT_TIMESTAMP_MAX_MIN_DIFF_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_TIMESTAMP_MAX_MIN_DIFF_ABS
    #: Mean peak maximum amplitude of events in packet (max over all base packets)
    MEAN_MAX = common_pb2.KpiMetricsEnum.KPI_MEAN_MAX
    #: Mean peak minimum amplitude of events in packet (min over all base packets)
    MEAN_MIN = common_pb2.KpiMetricsEnum.KPI_MEAN_MIN
    #: Mean peak absolute maximum amplitude of events in packet (max over all base packets)
    MEAN_MAX_ABS = common_pb2.KpiMetricsEnum.KPI_MEAN_MAX_ABS
    #: Mean peak absolute amplitude of events in packet (max over all base packets)
    EVENT_MEAN_MAX_MIN_DIFF_ABS = common_pb2.KpiMetricsEnum.KPI_EVENT_MEAN_MAX_MIN_DIFF_ABS
    #: Absolute difference between maximum and minimum amplitudes of events in packet, amplified
    MAX_MIN_DIFF_ABS_AMPLIFIED = common_pb2.KpiMetricsEnum.KPI_MAX_MIN_DIFF_ABS_AMPLIFIED


class Port(GrpcEnum):
    """
    An enumeration of port types

    Example:
        >>> port = Port.A

    Valid values are as follows:
    """
    A = common_pb2.Port.A  #: port A
    B = common_pb2.Port.B  #: port B
    C = common_pb2.Port.C  #: port C
    D = common_pb2.Port.D  #: port D
    E = common_pb2.Port.E  #: port E
    F = common_pb2.Port.F  #: port F
    G = common_pb2.Port.G  #: port G
    H = common_pb2.Port.H  #: port H


class SystemMode(GrpcEnum):
    """
    An enumeration of system modes

    Example:
        >>> sys_mode = SystemMode.SMARTBOX_PRO

    Valid values are as follows:
    """

    SMARTBOX_PRO = allegoserver_pb2.BackboneMode.SMARTBOX_PRO  #: Smartbox Pro
    #: Smartbox Pro Sinaps 256
    SMARTBOX_PRO_SINAPS_256 = allegoserver_pb2.BackboneMode.SMARTBOX_PRO_SINAPS_256
    #: Smartbox Pro Sinaps 512
    SMARTBOX_PRO_SINAPS_512 = allegoserver_pb2.BackboneMode.SMARTBOX_PRO_SINAPS_512
    #: Smartbox Pro Sinaps 1024
    SMARTBOX_PRO_SINAPS_1024 = allegoserver_pb2.BackboneMode.SMARTBOX_PRO_SINAPS_1024
    SMARTBOX_CLASSIC = allegoserver_pb2.BackboneMode.SMARTBOX_CLASSIC  #: Smartbox Classic
    #: Smartbox Sim Gen Sine
    SMARTBOX_SIM_GEN_SINE = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SINE
    #: Smartbox Sim Gen Sine Mapped
    SMARTBOX_SIM_GEN_SINE_MAPPED = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SINE_MAPPED
    #: Smartbox Sim Gen Sine High Freq
    SMARTBOX_SIM_GEN_SINE_HIGH_FREQ = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ
    #: Smartbox Sim Gen Sine Multi Band
    SMARTBOX_SIM_GEN_SINE_MULTI_BAND = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SINE_MULTI_BAND
    #: Smartbox Sim Gen Spikes
    SMARTBOX_SIM_GEN_SPIKES = allegoserver_pb2.BackboneMode.SMARTBOX_SIM_GEN_SPIKES
    OPEN_EPHYS_USB2 = allegoserver_pb2.BackboneMode.OPEN_EPHYS_USB2  #: Open Ephys USB2
    OPEN_EPHYS_USB3 = allegoserver_pb2.BackboneMode.OPEN_EPHYS_USB3  #: Open Ephys USB3
    INTAN_USB2 = allegoserver_pb2.BackboneMode.INTAN_USB2  #: Intan USB2
    #: Intan Recording Controller 1024
    INTAN_RECORDING_CONTROLLER_1024 = allegoserver_pb2.BackboneMode.INTAN_RECORDING_CONTROLLER_1024
    #: Intan Recording Controller 512
    INTAN_RECORDING_CONTROLLER_512 = allegoserver_pb2.BackboneMode.INTAN_RECORDING_CONTROLLER_512
    XDAQ_ONE_REC = allegoserver_pb2.BackboneMode.XDAQ_ONE_REC  #: XDAQ One Rec
    XDAQ_ONE_STIM = allegoserver_pb2.BackboneMode.XDAQ_ONE_STIM  #: XDAQ One Stim
    XDAQ_CORE_REC = allegoserver_pb2.BackboneMode.XDAQ_CORE_REC  #: XDAQ Core Rec
    XDAQ_CORE_STIM = allegoserver_pb2.BackboneMode.XDAQ_CORE_STIM  #: XDAQ Core Stim


class SignalType(GrpcEnum):
    """
    An enumeration of signal types

    Example:
        >>> sig_type = SignalType.AMP

    Valid values are as follows:
    """

    AMP = common_pb2.SignalType.PRI  #: primary signals
    AIN = common_pb2.SignalType.AUX  #: analog in
    DIN = common_pb2.SignalType.DIN  #: digital in
    DOUT = common_pb2.SignalType.DOUT  #: digital out



import getpass
import hashlib
import json
import os
import platform
import re
import subprocess
import sys
from pathlib import Path

import boto3
import botocore.exceptions
import radiens.utils.config as cfg
from grpc_interceptor import ClientCallDetails, ClientInterceptor


class MetadataClientInterceptor(ClientInterceptor):
    def __init__(self, id_token=None, device_uuid=None):
        self._id_token = id_token
        self._device_uuid = device_uuid

    def intercept(self, method, request, call_details):
        authorization = None
        if self._id_token is not None:
            authorization = f"Bearer {self._id_token}"
        elif self._device_uuid is not None:
            authorization = f"deviceUUID {self._device_uuid}"

        # replace the authorization metadata if it exists
        new_metadata = list(
            call_details.metadata) if call_details.metadata else []
        new_metadata = [(key, value) for key,
                        value in new_metadata if key.lower() != "authorization"]
        new_metadata.append(("authorization", authorization))

        new_details = ClientCallDetails(
            call_details.method,
            call_details.timeout,
            new_metadata,
            call_details.credentials,
            call_details.wait_for_ready,
            call_details.compression,
        )

        return method(request, new_details)


def get_hardware_uuid() -> str:
    """Retrieve the hardware UUID for macOS, Windows, and Linux."""
    system = platform.system()

    if system == "Darwin":
        return get_macos_hardware_uuid()
    elif system == "Windows":
        return get_windows_hardware_uuid()
    else:
        print(f"Unsupported OS: {system}", file=sys.stderr)
        return ""


def get_macos_hardware_uuid() -> str:
    """Retrieve the hardware UUID on macOS."""
    try:
        output = subprocess.check_output(
            ["ioreg", "-d2", "-c", "IOPlatformExpertDevice"], encoding="utf-8"
        )
    except subprocess.CalledProcessError as e:
        print("Failed to run ioreg:", e, file=sys.stderr)
        return ""

    match = re.search(r'"IOPlatformUUID"\s=\s"([^"]+)"', output)
    return match.group(1) if match else ""


def run_wmic_command(command: list) -> str:
    """Run a WMIC command and extract the relevant value."""
    try:
        output = subprocess.check_output(
            command, encoding="utf-8", stderr=subprocess.DEVNULL)
        lines = output.strip().split()
        if len(lines) > 1:
            return lines[1].strip()
    except subprocess.CalledProcessError:
        pass
    return ""


def get_windows_hardware_uuid() -> str:
    uuID = run_wmic_command(["wmic", "csproduct", "get", "UUID"])
    bios_serial = run_wmic_command(["wmic", "bios", "get", "serialnumber"])
    harddrive_serial = run_wmic_command(
        ["wmic", "DISKDRIVE", "get", "SerialNumber"])

    combined = (uuID + bios_serial + harddrive_serial).encode('utf-8')
    md5_hash = hashlib.md5(combined).hexdigest()
    return f"{md5_hash[:8]}-{md5_hash[8:12]}-{md5_hash[12:16]}-{md5_hash[16:20]}-{md5_hash[20:]}"


class SessionMetaData():
    def __init__(self):
        self._id_token = None
        self._hardware_uid = None
        is_offline = os.getenv("RADIENS_OFFLINE") == "1"
        if not is_offline:
            try:
                app_dir = cfg.get_user_app_data_dir()
                client_id = cfg.get_app_client_id()
                refresh_token, device_key = load_session(client_id, app_dir)
                self._id_token = get_id_token(refresh_token, device_key)
            except botocore.exceptions.EndpointConnectionError as e:
                is_offline = True
            except botocore.exceptions.ClientError as e:
                if e.response['ResponseMetadata']['HTTPStatusCode'] == 400:
                    is_offline = True

        if is_offline:
            self._hardware_uid = get_hardware_uuid()


# ====== helpers ======
def get_id_token(refresh_token, device_key):
    if refresh_token is None or device_key is None:
        return get_id_token_password()
    if refresh_token is not None and device_key is not None:
        return get_id_token_refresh(refresh_token, device_key)
    return None


def get_id_token_refresh(refresh_token, device_key):
    client = boto3.client('cognito-idp', region_name=cfg.REGION)
    auth_resp = client.initiate_auth(
        AuthFlow='REFRESH_TOKEN_AUTH',
        AuthParameters={
            'REFRESH_TOKEN': refresh_token,
            'DEVICE_KEY': device_key
        },
        ClientId=cfg.get_app_client_id()
    )

    client.close()
    return auth_resp['AuthenticationResult']['IdToken']


def get_id_token_password():
    _email = input('Enter user email address: ')
    _password = getpass.getpass(prompt='Enter use password: ')

    client = boto3.client('cognito-idp', region_name=cfg.REGION)
    auth_resp = client.initiate_auth(
        AuthFlow='USER_PASSWORD_AUTH',
        AuthParameters={
            'USERNAME': _email,
            'PASSWORD': _password
        },
        ClientId=cfg.get_app_client_id()
    )
    client.close()

    return auth_resp['AuthenticationResult']['IdToken']


def load_session(client_id, app_dir):
    file = Path(app_dir, "session.json")
    if not file.is_file() or file.stat().st_size == 0:
        return None, None
    with open(file) as fid:
        session_info = json.load(fid)
        if 'CognitoIdentityServiceProvider' in session_info.keys():
            session_info = session_info['CognitoIdentityServiceProvider']
        else:
            return None, None
    refresh_token, device_key = (None, None)
    for _client_id, _app_client in session_info.items():
        if _client_id == client_id:
            if 'LastAuthUser' in _app_client.keys():
                user_key = _app_client['LastAuthUser']
                user_body = _app_client[user_key]
                fields = user_body.keys()
                refresh_token = user_body['refreshToken'] if 'refreshToken' in fields else None
                device_key = user_body['deviceKey'] if 'deviceKey' in fields else None

                break

    return refresh_token, device_key



import datetime
import json

import pandas as pd
from radiens.api.api_utils.util import ext_to_radiens_file_type, to_file_ext
from radiens.grpc_radiens import allegoserver_pb2, common_pb2, datasource_pb2
from radiens.lib.channel_metadata import ChannelMetadata
from radiens.utils.constants import TimeRange
from radiens.utils.enums import RadiensFileType, SignalType
from radiens.utils.util import generic_repr, make_time_range


class FileSetDescriptor():
    '''
    Radiens data file descriptor.
    '''

    def __init__(self, raw_msg=None):
        """
        """
        self._path = None
        self._base_name = None
        self._file_type = None
        self._file_uid = None
        self._dsource_idx = 0
        self._timestamp = ''
        if not raw_msg is None:
            self._path = raw_msg.path
            self._base_name = raw_msg.baseName
            self._file_type = to_file_ext(raw_msg.fileType)
            self._file_uid = raw_msg.fileNameUID

    @property
    def path(self) -> str:
        """
        Path to Radiens data file set.
        """
        return self._path

    @property
    def base_name(self) -> str:
        """
        Base name of Radiens data file set.
        """
        return self._base_name

    @property
    def file_type(self) -> str:
        """
        File type of Radiens data file set.
        """
        return self._file_type

    @property
    def file_uid(self) -> any:
        """
        File set UID (optional). Returns None if no index will be appended
        """
        return self._file_uid

    @property
    def file_name_index(self) -> any:
        """
        Index appended to the base file name (optional). Returns None if no index will be appended.
        """
        return self._dsource_idx

    @property
    def file_name_timestamp(self) -> any:
        """
        Returns True if timestamp is appended to the file name.
        """
        return self._timestamp

    def set_from_rec_cfg(self, msg: allegoserver_pb2.ConfigRecording):
        self._path = msg.baseFilePath
        self._base_name = msg.baseFileName
        self._file_type = to_file_ext(common_pb2.XDAT)
        self._dsource_idx = msg.dataSourceIdx
        self._timestamp = msg.timeStamp


class FileSetStat():
    '''
    Radiens data file set summary status.
    '''

    def __init__(self, raw_msg: datasource_pb2.DataSourceFileSetStat):
        """
        """
        self._num_files = raw_msg.numFiles
        self._num_bytes_primary_data = raw_msg.numBytesPrimaryData
        self._num_bytes_meta_data = raw_msg.numBytesMetaData
        self._num_bytes_total = raw_msg.numBytes
        self._is_metadata = raw_msg.isMetadataFile
        self._timestamp = raw_msg.timeStamp.ToDatetime()
        self._num_chan = raw_msg.numChannels
        self._dur_sec = raw_msg.durationSec
        self._dset_spec = {'uid': raw_msg.datasetUID,
                           'checksum': raw_msg.datasetChecksum, 'provenance': []}
        for prov in raw_msg.datasetProvenance:
            self._dset_spec['provenance'].append(prov)
        self._fs = raw_msg.sampleRate
        self._background_proc = None
        if raw_msg.backgroundProc is not None:
            self._background_proc = {'id': raw_msg.backgroundProc.uID,
                                     'frac_complete': raw_msg.backgroundProc.fracComplete,
                                     'elapsed_time_sec': raw_msg.backgroundProc.elapsedTimeSec,
                                     'est_time_completion_walltime': raw_msg.backgroundProc.estimatedCompletionWallTime}

    @ property
    def num_files(self) -> int:
        """
        Number of files in the file set.
        """
        return self._num_files

    @ property
    def num_bytes(self) -> dict:
        """
        Number of bytes in file set, keys are "primary_data", "meta_data", and "total"
        """
        return {'primary_data': self._num_bytes_primary_data, 'meta_data': self._num_bytes_meta_data, 'total': self._num_bytes_total}

    @ property
    def has_meta_data_file(self) -> bool:
        """
        Flag for existence of Radiens meta data file.
        """
        return self._is_metadata

    @ property
    def timestamp(self) -> datetime.datetime:
        """
        File set time stamp.
        """
        return self._timestamp

    @ property
    def num_chan(self) -> int:
        '''
        Total number of channels in the data file of all signal types, i.e., all amp, gpio-ain, gpio-din, and gpio-dout channels.
        '''
        return self._num_chan

    @ property
    def dur_sec(self) -> float:
        '''
        Duration of file set in seconds.
        '''
        return self._dur_sec

    @ property
    def dataset_spec(self) -> dict:
        '''
        Dataset specifications, with keys = "uid", "checksum", "provenance"
        '''
        return self._dset_spec

    @ property
    def samp_freq(self) -> float:
        '''
        Dataset sample frequency in samples/sec.
        '''
        return self._fs

    @ property
    def fs(self) -> float:
        '''
        Aliaes `dataset.samp_freq`.
        '''
        return self._fs

    @ property
    def background_proc(self) -> dict:
        '''
        Background processing specifications (system reserved)
        '''
        return self._background_proc


class DatasetFileSet():
    '''
    '''

    def __init__(self, raw_msg: datasource_pb2.CpRmMvLsReply.DataSourceInfo):
        """
        Summary information on the file set of one Radiens dataset.
        """
        self._desc = FileSetDescriptor(raw_msg.desc)
        self._stat = FileSetStat(raw_msg.stat)

    @ property
    def descriptor(self) -> FileSetDescriptor:
        """
        File set descriptor providing path, base name, and file type.
        """
        return self._desc

    @ property
    def stat(self) -> FileSetStat:
        """
        File set summary status including duration, number of channels, etc.
        """
        return self._stat


class DatasetMetadata():
    '''
    '''

    def __init__(self, raw_msg: datasource_pb2.DataSourceSetSaveReply = None):
        """
            Meta data describing one Radiens data file or dataset.
        """
        self._d = {'dsource_id': '',
                   'kpi_dsource_id': '',
                   'TR': make_time_range(time_range=[0, 0], fs=1.0),
                   'channel_metadata': ChannelMetadata(),
                   'dataset_uid': '',
                   'source_type': '',
                   'file_type': '',
                   'source_mode': '',
                   'label': '',
                   'path': '',
                   'base_name': ''}
        if raw_msg is None:
            return

        if len(raw_msg.status.tR.timestamp) == 2 and len(raw_msg.status.tR.timeRangeSec) == 2:
            tr = make_time_range(pbTR=raw_msg.status.tR)
        else:  # deprecated, but for backward compatibility atm
            tr = make_time_range(timestamp=list(raw_msg.status.timestampRange),
                                 time_range=list(raw_msg.status.timeRange),
                                 fs=float(raw_msg.status.sampleFreq))

        self._d = {'dsource_id': raw_msg.dsourceID,
                   'kpi_dsource_id': raw_msg.status.kpiDsourceID,
                   'TR': tr,
                   'channel_metadata': ChannelMetadata(raw_msg.status.signalGroup),
                   'dataset_uid': raw_msg.status.uid,
                   'source_type': raw_msg.status.dataSourceType,
                   'file_type': RadiensFileType.parse(raw_msg.status.fileType),
                   'source_mode': raw_msg.status.mode,
                   'label': raw_msg.status.label,
                   'path': raw_msg.status.path,
                   'base_name': raw_msg.status.baseFileName}

    @ property
    def id(self) -> str:
        """
        Dataset ID of this DataSource the Radiens hub.  `None` indicates that it is not on a hub.
        """
        return self._d['dsource_id']

    @ property
    def time_range(self) -> TimeRange:
        """
        Dataset time range (:py:class:`~radiens.utils.constants.TimeRange`)
        """
        return self._d['TR']

    @ property
    def TR(self) -> TimeRange:
        """
        Alias for `time_range`.
        """
        return self._d['TR']

    @ property
    def channel_metadata(self) -> ChannelMetadata:
        """
        Dataset's ChannelMetadata, which describes the dataset's signals (aka channels), probes, positions, etc.
        """
        return self._d['channel_metadata']

    @ property
    def attributes(self) -> dict:
        """
        Dataset attributes.
        """
        return self._d

    @ property
    def path(self) -> str:
        """
        File system path of the Radiens data file set that backs this DataSource.
        """
        return self._d['path']

    @ property
    def base_name(self) -> str:
        """
        File system base name of the Radiens data file set that backs this DataSource.
        """
        return self._d['base_name']

    @ property
    def file_type(self) -> RadiensFileType:
        """
        File type of the Radiens data file set that backs this DataSource.
        """
        return self._d['file_type']

    @ property
    def table(self) -> pd.DataFrame:
        """
        Dataset attributes as a table.
        """
        return pd.DataFrame({'path': [self.path],
                             'name': [self.base_name],
                             'type': [self._d['file_type'].name],
                             'time range (sec)': ['[{:.3f}, {:.3f}]'.format(*self.time_range.sec)],
                             'duration (sec)': ['{:.3f}'.format(self.time_range.dur_sec)],
                             'channels': [self.channel_metadata.num_sigs(SignalType.AMP)],
                             'gpio': [[self.channel_metadata.num_sigs(SignalType.AIN), self.channel_metadata.num_sigs(SignalType.DIN), self.channel_metadata.num_sigs(SignalType.DOUT)]],
                             'sample freq': ['{:.0f}'.format(self.time_range.fs)],
                             'dataset UID': [self._d['dataset_uid']],
                             'label': [self._d['label']],
                             'hub ID (mutable)': [self.id]}).T.rename({0: 'Value'}, axis='columns')

    def __str__(self) -> str:
        return self.table.__str__()

    def __repr__(self) -> str:
        return self.table.__repr__()

    def clear_dataset_id(self) -> None:
        """
        Clears this DataSource's dataset ID by setting it to None.
        This is a power user function. Best practice is to call it when the dataset has been cleared from the Radiens hub.

        """
        self._d['dataset_id'] = None
